# pragma: no cover

# stdlib
import sys

if not bool(getattr(sys, "ps1", sys.flags.interactive)):

	try:
		# stdlib
		import readline

		# typeshed thinks the module but not these functions are available on Windows.
		# In reality the whole module is unavailable.
		readline.set_history_length(0)
		readline.set_auto_history(False)
	except (ImportError, AttributeError):
		# Attribute error on PyPy, ImportError on Windows etc.
		pass



#!/usr/bin/env python3
#
#  tracebacks.py
"""
Functions for handling exceptions and their tracebacks.

.. versionadded:: 1.0.0
.. latex:vspace:: -10px
"""
#
#  Copyright © 2020-2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import contextlib
import sys
from typing import TYPE_CHECKING, Callable, ContextManager, List, Type, TypeVar, Union

# 3rd party
import click
from domdf_python_tools.compat import nullcontext

if sys.version_info >= (3, 7) or TYPE_CHECKING:
	# stdlib
	from typing import NoReturn

__all__ = ("TracebackHandler", "handle_tracebacks", "traceback_handler", "traceback_option")

_C = TypeVar("_C", bound=click.Command)


class TracebackHandler:
	"""
	Context manager to abort execution with a short error message
	on the following exception types:

	* :exc:`FileNotFoundError`
	* :exc:`FileExistsError`

	Other custom exception classes inheriting from :exc:`Exception` are also handled,
	but with a generic message.

	The following exception classes are ignored:

	* :exc:`EOFError`
	* :exc:`KeyboardInterrupt`
	* :exc:`click.ClickException`
	* :exc:`SystemExit` (new in version 1.1.2)

	How these exceptions are handled can be changed, and supported can be added for
	further exception classes by subclassing this class.
	Each method is named in the form :file:`handle_{<exception>}`, where ``exception``
	is the name of the exception class to handle.

	.. versionadded:: 1.0.0

	:param exception: The exception to raise after handling the traceback.
		If not running within a click command or group you'll likely
		want to set this to :exc:`SystemExit(1) <SystemExit>`.
	:default exception: :class:`click.Abort() <click.Abort>`

	.. versionchanged:: 1.4.0  Added the ``exception`` argument.

	.. seealso:: :func:`~.handle_tracebacks`.
	.. latex:clearpage::
	"""  # noqa: D400

	exception: Exception
	"""
	The exception to raise after handling the traceback.

	.. versionadded:: 1.4.0
	"""

	def __init__(self, exception: BaseException = click.Abort()):
		self.exception: BaseException = exception  # type: ignore[assignment]

	def abort(self, msg: Union[str, List[str]]) -> "NoReturn":
		"""
		Abort the current process by calling ``self.exception``.

		.. versionadded:: 1.4.0

		:param msg: The message to write to stderr before raising the exception.
			If a list of strings the strings are concatenated (i.e. ``''.join(msg)``).
		"""

		if not isinstance(msg, str):
			msg = ''.join(msg)

		if msg:
			click.echo(msg, err=True, color=False)

		raise self.exception

	def handle_EOFError(self, e: EOFError) -> "NoReturn":  # noqa: D102
		raise e

	def handle_KeyboardInterrupt(self, e: KeyboardInterrupt) -> "NoReturn":  # noqa: D102
		raise e

	def handle_ClickException(self, e: click.ClickException) -> "NoReturn":  # noqa: D102
		raise e

	def handle_Abort(self, e: click.Abort) -> "NoReturn":  # noqa: D102
		raise e

	def handle_SystemExit(self, e: SystemExit) -> "NoReturn":  # noqa: D102
		raise e

	def handle_FileNotFoundError(self, e: FileNotFoundError) -> "NoReturn":  # noqa: D102
		self.abort(f"File Not Found: {e}")

	def handle_FileExistsError(self, e: FileExistsError) -> "NoReturn":  # noqa: D102
		self.abort(f"File Exists: {e}")

	def handle(self, e: BaseException) -> bool:
		"""
		Handle the given exception.

		:param e:
		"""

		exception_name = e.__class__.__name__

		if hasattr(self, f"handle_{exception_name}"):
			return getattr(self, f"handle_{exception_name}")(e)

		for base in e.__class__.__mro__:
			if hasattr(self, f"handle_{base.__name__}"):
				return getattr(self, f"handle_{base.__name__}")(e)

		self.abort(f"An error occurred: {e}")

	@contextlib.contextmanager
	def __call__(self):  # noqa: MAN002
		"""
		Use the :class:`~.TracebackHandler` with a :keyword:`with` block, and handle any exceptions raised within.
		"""

		try:
			yield
		except BaseException as e:
			self.handle(e)


@contextlib.contextmanager
def traceback_handler():  # noqa: MAN002
	"""
	Context manager to abort execution with a short error message on the following exception types:

	* :exc:`FileNotFoundError`
	* :exc:`FileExistsError`

	Other custom exception classes inheriting from :exc:`Exception` are also handled,
	but with a generic message.

	The following exception classes are ignored:

	* :exc:`EOFError`
	* :exc:`KeyboardInterrupt`
	* :exc:`click.ClickException`

	.. versionadded:: 0.8.0

	.. seealso:: :func:`~.handle_tracebacks` and :class:`~.TracebackHandler`
	"""  # noqa: D400

	with TracebackHandler()():
		yield


def handle_tracebacks(
		show_traceback: bool = False,
		cls: Type[TracebackHandler] = TracebackHandler,
		) -> ContextManager:
	"""
	Context manager to conditionally handle tracebacks, usually based on the value of a command line flag.

	.. versionadded:: 0.8.0

	:param show_traceback: If :py:obj:`True`, the full Python traceback will be shown on errors.
		If :py:obj:`False`, only the summary of the traceback will be shown.
		In either case the program execution will stop on error.
	:param cls: The class to use to handle the tracebacks.

	:rtype:

	.. versionchanged:: 1.0.0  Added the ``cls`` parameter.
	.. seealso:: :func:`~.traceback_handler` and :class:`~.TracebackHandler`
	"""

	if show_traceback:
		return nullcontext()
	else:
		return cls()()


def traceback_option(help_text: str = "Show the complete traceback on error.") -> Callable[[_C], _C]:
	"""
	Decorator to add the ``-T / --traceback`` option to a click command.

	The value is exposed via the parameter ``show_traceback``: :class:`bool`.

	.. versionadded:: 1.0.0

	:param help_text: The help text for the option.
	"""

	# this package
	from consolekit.options import flag_option

	return flag_option("-T", "--traceback", "show_traceback", help=help_text)



#!/usr/bin/env python3
#
#  input.py
"""
Input functions (prompt, choice etc.).
"""
#
#  Copyright © 2020-2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  prompt and confirm based on https://github.com/pallets/click
#  Copyright 2014 Pallets
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |      * Redistributions of source code must retain the above copyright notice,
#  |        this list of conditions and the following disclaimer.
#  |      * Redistributions in binary form must reproduce the above copyright notice,
#  |        this list of conditions and the following disclaimer in the documentation
#  |        and/or other materials provided with the distribution.
#  |      * Neither the name of the copyright holder nor the names of its contributors
#  |        may be used to endorse or promote products derived from this software without
#  |        specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  |  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  |  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  |  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
#  |  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
#  |  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#  |  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
#  |  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
#  |  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
#  |  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#  |
#
#  stderr_input based on raw_input from https://foss.heptapod.net/pypy/pypy
#  PyPy Copyright holders 2003-2020
#  MIT Licenced
#

# stdlib
import sys
from typing import IO, Any, Callable, List, Mapping, Optional, Union, overload

# 3rd party
import click
from click.termui import _build_prompt, hidden_prompt_func
from click.types import Path, convert_type

# this package
from consolekit import _readline  # noqa: F401
from consolekit._types import _ConvertibleType

__all__ = (
		"prompt",
		"confirm",
		"stderr_input",
		"choice",
		)


def prompt(  # noqa: MAN002
		text: str,
		default: Optional[str] = None,
		hide_input: bool = False,
		confirmation_prompt: Union[bool, str] = False,
		type: Optional[_ConvertibleType] = None,  # noqa: A002  # pylint: disable=redefined-builtin
		value_proc: Optional[Callable[[Optional[str]], Any]] = None,
		prompt_suffix: str = ": ",
		show_default: bool = True,
		err: bool = False,
		show_choices: bool = True,
		):
	"""
	Prompts a user for input.

	If the user aborts the input by sending an interrupt signal,
	this function will catch it and raise a :exc:`click.Abort` exception.

	:param text: The text to show for the prompt.
	:param default: The default value to use if no input happens.
		If this is not given it will prompt until it is aborted.
	:param hide_input: If :py:obj:`True` then the input value will be hidden.
	:param confirmation_prompt: Asks for confirmation for the value.
		Can be set to a string instead of :py:obj:`True` to customize the message.
	:param type: The type to check the value against.
	:param value_proc: If this parameter is provided it must be a function that
		is invoked instead of the type conversion to convert a value.
	:param prompt_suffix: A suffix that should be added to the prompt.
	:param show_default: Shows or hides the default value in the prompt.
	:param err: If :py:obj:`True` the file defaults to ``stderr`` instead of
		``stdout``, the same as with :func:`click.echo`.
	:param show_choices: Show or hide choices if the passed type is a :class:`click.Choice`.
		For example, if the choice is either ``day`` or ``week``,
		``show_choices`` is :py:obj:`True` and ``text`` is ``'Group by'`` then the
		prompt will be ``'Group by (day, week): '``.
	"""

	result = None  # noqa

	def prompt_func(text: Any) -> Any:
		try:
			return _prompt(text, err=err, hide_input=hide_input)
		except (KeyboardInterrupt, EOFError):
			if hide_input:
				click.echo(None, err=err)
			raise click.Abort()

	if value_proc is None:
		value_proc = convert_type(type, default)

	prompt = _build_prompt(
			text,
			prompt_suffix,
			show_default,
			default,
			show_choices,
			type,  # type: ignore[arg-type]
			)

	has_default = default is not None

	while True:
		while True:
			value = prompt_func(prompt)

			if value:
				break
			elif has_default:
				if isinstance(value_proc, Path):  # pylint: disable=loop-invariant-statement
					# validate Path default value (exists, dir_okay etc.)
					value = default
					break
				return default

		try:  # pylint: disable=loop-try-except-usage
			result = value_proc(value)
		except click.UsageError as e:
			click.echo(f"Error: {e.message}", err=err)  # pylint: disable=loop-invariant-statement
			continue

		if not confirmation_prompt:
			return result

		if confirmation_prompt is True:
			confirmation_prompt = "Repeat for confirmation: "

		while True:
			value2 = prompt_func(confirmation_prompt)
			if value2:
				break

		if value == value2:  # pylint: disable=loop-invariant-statement
			return result

		click.echo("Error: the two entered values do not match", err=err)


def confirm(  # noqa: MAN002
		text: str,
		default: bool = False,
		abort: bool = False,
		prompt_suffix: str = ": ",
		show_default: bool = True,
		err: bool = False,
		):
	"""
	Prompts for confirmation (yes/no question).

	If the user aborts the input by sending a interrupt signal this
	function will catch it and raise a :exc:`click.Abort` exception.

	.. latex:clearpage::

	:param text: The question to ask.
	:param default: The default for the prompt.
	:param abort: If :py:obj:`True` a negative answer aborts the exception by raising :exc:`click.Abort`.
	:param prompt_suffix: A suffix that should be added to the prompt.
	:param show_default: Shows or hides the default value in the prompt.
	:param err: If :py:obj:`True` the file defaults to ``stderr`` instead of ``stdout``, the same as with echo.
	"""

	prompt = _build_prompt(text, prompt_suffix, show_default, "Y/n" if default else "y/N")

	while True:
		try:  # pylint: disable=loop-try-except-usage
			value = _prompt(prompt, err=err, hide_input=False).lower().strip()
		except (KeyboardInterrupt, EOFError):
			raise click.Abort()

		if value in ('y', "yes"):
			rv = True
		elif value in ('n', "no"):
			rv = False
		elif value == '':
			rv = default
		else:
			click.echo("Error: invalid input", err=err)
			continue
		break

	if abort and not rv:
		raise click.Abort()

	return rv


def stderr_input(prompt: str = '', file: IO = sys.stdout) -> str:  # pragma: no cover
	"""
	Read a string from standard input, but prompt to standard error.

	The trailing newline is stripped.
	If the user hits EOF (Unix: :kbd:`Ctrl-D`, Windows: :kbd:`Ctrl-Z+Return`), raise :exc:`EOFError`.

	On Unix, GNU readline is used if enabled.

	The ``prompt`` string, if given, is printed to stderr without a trailing newline before reading.
	"""

	if file is sys.stdout:
		return input(prompt)

	try:
		stdin = sys.stdin
	except AttributeError:
		raise RuntimeError("stderr_input: lost sys.stdin")

	file.write(prompt)

	try:
		flush = file.flush
	except AttributeError:
		pass
	else:
		flush()

	try:
		file.softspace = 0  # type: ignore[attr-defined]
	except (AttributeError, TypeError):
		pass

	line = stdin.readline()

	if not line:  # inputting an empty line gives line == '\n'
		raise EOFError
	elif line[-1] == '\n':
		return line[:-1]

	return line


def _prompt(text: Any, err: bool, hide_input: bool):  # noqa: MAN002
	if sys.platform != "linux":
		# Write the prompt separately so that we get nice
		# coloring through colorama on Windows
		click.echo(text, nl=False, err=err)
		text = ''

	if hide_input:
		return hidden_prompt_func(text)
	elif err:
		return stderr_input(text, file=sys.stderr)
	else:
		return click.termui.visible_prompt_func(text)


@overload
def choice(
		options: List[str],
		text: str = ...,
		default: Optional[str] = ...,
		prompt_suffix: str = ...,
		show_default: bool = ...,
		err: bool = ...,
		start_index: int = ...
		) -> int: ...


@overload
def choice(
		options: Mapping[str, str],
		text: str = ...,
		default: Optional[str] = ...,
		prompt_suffix: str = ...,
		show_default: bool = ...,
		err: bool = ...,
		start_index: int = ...
		) -> str: ...


def choice(
		options: Union[List[str], Mapping[str, str]],
		text: str = '',
		default: Optional[str] = None,
		prompt_suffix: str = ": ",
		show_default: bool = True,
		err: bool = False,
		start_index: int = 0
		) -> Union[str, int]:
	"""
	Prompts a user for input.

	If the user aborts the input by sending an interrupt signal, this
	function will catch it and raise a :exc:`click.Abort` exception.

	:param options:
	:param text: The text to show for the prompt.
	:param default: The index of the default value to use if the user does not enter anything.
		If this is not given it will prompt the user until aborted.
	:param prompt_suffix: A suffix that should be added to the prompt.
	:param show_default: Shows or hides the default value in the prompt.
	:param err: If :py:obj:`True` the file defaults to ``stderr`` instead of
		``stdout``, the same as with echo.
	:param start_index: If ``options`` is a list of values, sets the start index.
	"""

	# TODO: completer for numbers?

	type_: click.ParamType

	if isinstance(options, Mapping):
		# (Y/I/N/O/D/Z) [default=N]

		text = f"{text} ({'/'.join(options.keys())})"
		type_ = click.STRING

		for choice, descripton in options.items():
			click.echo(f" {choice} : {descripton}")

	else:
		type_ = click.IntRange(start_index, len(options) + 1 - start_index)

		for idx, descripton in enumerate(options):
			idx += start_index
			click.echo(f" [{idx}] {descripton}")

	if default is not None and show_default:
		text += f" [default={default}]"

	while True:
		selection = prompt(
				text=text,
				default=default,
				type=type_,
				prompt_suffix=prompt_suffix,
				show_default=False,
				err=err,
				)
		# pylint: disable=loop-invariant-statement
		if isinstance(options, Mapping):
			selection = selection.strip().upper()
			if selection not in options:
				click.echo("Please enter a valid option.")
			else:
				return selection
		else:
			return selection - start_index
		# pylint: enable=loop-invariant-statement



#!/usr/bin/env python
#
#  terminal_colours.py
"""
Functions for adding ANSI character codes to terminal print statements.

.. seealso:: http://en.wikipedia.org/wiki/ANSI_escape_code
"""
#
#  Copyright © 2020-2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  Based on colorama
#  https://github.com/tartley/colorama
#  Copyright Jonathan Hartley 2013
#  Distributed under the BSD 3-Clause license.
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  * Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |  * Redistributions in binary form must reproduce the above copyright notice,
#  |    this list of conditions and the following disclaimer in the documentation
#  |    and/or other materials provided with the distribution.
#  |
#  |  * Neither the name of the copyright holders, nor those of its contributors
#  |    may be used to endorse or promote products derived from this software without
#  |    specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  |  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  |  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  |  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  |  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  |  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  |  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#  Includes modifications to colorama made by Bram Geelen in
#  https://github.com/tartley/colorama/pull/141/files
#
#  resolve_color_default, _ansi_re and strip_ansi based on
#  https://github.com/pallets/click
#  Copyright 2014 Pallets
#  Distributed under the BSD 3-Clause license.
#

# stdlib
import os
import re
from abc import ABC
from collections import deque
from functools import partial
from typing import Deque, Iterable, List, Optional, Pattern, Type, TypeVar, Union

# 3rd party
import click
from typing_extensions import Final

__all__ = (
		"AnsiFore",
		"AnsiBack",
		"AnsiStyle",
		"AnsiCursor",
		"Fore",
		"Back",
		"Style",
		"Cursor",
		"AnsiCodes",
		"Colour",
		"ColourTrilean",
		"clear_line",
		"code_to_chars",
		"resolve_color_default",
		"set_title",
		"strip_ansi",
		"CSI",
		"OSC",
		"BEL",
		)

try:
	# 3rd party
	import colorama  # type: ignore[import]
	colorama.init()

except ImportError:
	pass

_C = TypeVar("_C", bound="Colour")
_AC = TypeVar("_AC", bound="AnsiCodes")

ColourTrilean = Optional[bool]
"""
Represents the :py:obj:`True`/:py:obj:`False`/:py:obj:`None` state of colour options.

.. versionadded:: 0.8.0
"""

CSI: Final[str] = "\u001b["
OSC: Final[str] = "\u001b]"
BEL: Final[str] = '\x07'

fore_stack: Deque[str] = deque()
back_stack: Deque[str] = deque()
style_stack: Deque[str] = deque()


def resolve_color_default(color: ColourTrilean = None) -> ColourTrilean:
	"""
	Helper to get the default value of the color flag.

	If a value is passed it is returned unchanged,
	otherwise it's looked up from the current context.

	If the environment variable ``PYCHARM_HOSTED`` is ``1``
	(which is the case if running in PyCharm)
	the output will be coloured by default.

	If the environment variable ``NO_COLOR`` is ``1``
	the output will not be coloured by default.
	See https://no-color.org/ for more details.
	This variable takes precedence over ``PYCHARM_HOSTED``.

	If no value is passed in, there is no context,
	and neither environment variable is set,
	:py:obj:`None` is returned.

	.. versionchanged:: 1.3.0

		* Added support for the ``NO_COLOR`` environment variable.
		* Only uses a value from the click context (:attr:`Context.color <click.Context.color>`)
		  if it is not :py:obj:`None`. Otherwise falls back to checking the environment variables.

	:param color:
	"""

	if color is not None:
		return color

	ctx = click.get_current_context(silent=True)

	if ctx is not None and ctx.color is not None:
		return ctx.color

	if int(os.environ.get("NO_COLOR", 0)):
		return False

	if int(os.environ.get("PYCHARM_HOSTED", 0)):
		return True

	return None


def code_to_chars(code: Union[str, int]) -> str:  # noqa: D103
	return CSI + str(code) + 'm'


def set_title(title: str) -> str:  # noqa: D103
	return OSC + "2;" + title + BEL


# def clear_screen(mode: int = 2) -> str:
# 	return CSI + str(mode) + 'J'


def clear_line(mode: int = 2) -> str:  # noqa: D103
	return CSI + str(mode) + 'K'


_ansi_re: Pattern[str] = re.compile(r"\033\[[;?0-9]*[a-zA-Z]")


def strip_ansi(value: str) -> str:
	"""
	Strip ANSI colour codes from the given string to return a plaintext output.

	:param value:
	"""

	return _ansi_re.sub('', value)


class Colour(str):
	r"""
	An ANSI escape sequence representing a colour.

	The colour can be used as a context manager, a string, or a function.

	:param style: Escape sequence representing the style.
	:type style: :class:`str`
	:param stack: The stack to place the escape sequence on.
	:type stack: :class:`~typing.List`\[:class:`str`\]
	:param reset: The escape sequence to reset the style.
	:type reset: :class:`str`

	.. autosummary-widths:: 7/16
	"""

	__slots__ = ("style", "reset", "stack")

	style: str
	reset: str
	stack: Union[Deque[str], List[str]]

	def __new__(cls, style: str, stack: Union[Deque[str], List[str]], reset: str) -> "Colour":  # noqa: D102
		self = super().__new__(cls, style)  # type: ignore[import]
		self.style = style
		self.stack = stack
		self.reset = reset

		return self

	def __enter__(self) -> None:
		print(self.style, end='')
		self.stack.append(self.style)

	def __exit__(self, exc_type, exc_val, exc_tb) -> None:
		if self.style == self.stack[-1]:
			self.stack.pop()
			print(self.stack[-1], end='')

	def __call__(self, text) -> str:  # noqa: MAN001
		"""
		Returns the given text in this colour.
		"""

		return f"{self}{text}{self.reset}"

	@classmethod
	def from_code(cls: Type[_C], code: Union[str, int], background: bool = False) -> _C:
		"""
		Returns a :class:`~.Colour` to create coloured text.

		The colour can be reset using :py:obj:`.Fore.RESET` or :py:obj:`.Back.RESET`.

		.. versionadded:: 0.9.0

		:param code: A 3- or 4- bit ANSI colour code.
		:param background: Whether to set the colour for the background.

		:rtype: :class:`~.Colour`

		.. note::

			The ``background`` option only influences the reset value and the stack used.
			It will not handle conversion of foreground codes to background codes.

		.. latex:clearpage::
		"""

		if background:
			stack = back_stack
			reset = AnsiBack._reset
		else:
			stack = fore_stack
			reset = AnsiFore._reset

		return cls(code_to_chars(code), stack=stack, reset=reset)

	@classmethod
	def from_256_code(cls: Type[_C], code: Union[str, int], background: bool = False) -> _C:
		"""
		Returns a :class:`~.Colour` to create 256-colour text.

		The colour can be reset using :py:obj:`.Fore.RESET` or :py:obj:`.Back.RESET`.

		.. versionadded:: 0.9.0
		.. note:: Not all terminals support 256-colour mode.

		:param code: A 256-colour ANSI code.
		:param background: Whether to set the colour for the background.

		:rtype: :class:`~.Colour`

		.. note::

			The ``background`` option only influences the reset value and the stack used.
			It will not handle conversion of foreground codes to background codes.
		"""

		if background:
			stack = back_stack
			reset = AnsiBack._reset
			template = CSI + "48;5;{}m"

		else:
			stack = fore_stack
			reset = AnsiFore._reset
			template = CSI + "38;5;{}m"

		return cls(template.format(code), stack=stack, reset=reset)

	@classmethod
	def from_rgb(
			cls: Type[_C],
			r: Union[str, int],
			g: Union[str, int],
			b: Union[str, int],
			background: bool = False
			) -> _C:
		"""
		Returns a :class:`~.Colour` to create 24-bit coloured text.

		The colour can be reset using :py:obj:`.Fore.RESET` or :py:obj:`.Back.RESET`.

		.. versionadded:: 0.9.0

		.. note:: Not all terminals support 24-bit colours.

		:param r:
		:param g:
		:param b:
		:param background: Whether to set the colour for the background.

		:rtype: :class:`~.Colour`

		.. note::

			The ``background`` option only influences the reset value and the stack used.
			It will not handle conversion of foreground codes to background codes.

		.. latex:clearpage::
		"""

		if background:
			template = CSI + "48;2;{r};{g};{b}m"
			stack = back_stack
			reset = AnsiBack._reset

		else:
			template = CSI + "38;2;{r};{g};{b}m"
			stack = fore_stack
			reset = AnsiFore._reset

		return cls(template.format(r=r, g=g, b=b), stack=stack, reset=reset)

	@classmethod
	def from_hex(cls: Type[_C], hex_colour: str, background: bool = False) -> _C:
		"""
		Returns a :class:`~.Colour` to create 24-bit coloured text.

		The colour can be reset using :py:obj:`.Fore.RESET` or :py:obj:`.Back.RESET`.

		.. versionadded:: 0.9.0

		.. note:: Not all terminals support 24-bit colours.

		:param hex_colour: The hex colour code.
		:param background: Whether to set the colour for the background.

		:rtype: :class:`~.Colour`

		.. note::

			The ``background`` option only influences the reset value and the stack used.
			It will not handle conversion of foreground codes to background codes.
		"""

		# From https://stackoverflow.com/q/29643352
		# https://stackoverflow.com/users/3924370/julian-white
		# CC BY-SA 3.0

		value = hex_colour.lstrip('#')
		lv = len(value)
		r, g, b = tuple(int(value[i:i + lv // 3], 16) for i in range(0, lv, lv // 3))

		return cls.from_rgb(r, g, b, background=background)


def print_256_colour_testpattern() -> None:
	"""
	Print a 256-colour test pattern to the terminal.

	.. versionadded:: 0.9.0

	.. note:: Not all terminals support 24-bit colours.
	"""

	# 3rd party
	from domdf_python_tools.iterative import chunks

	def print_heading(text: str, block_size: int = 3, n_blocks: int = 36) -> None:
		click.echo()
		click.echo(str(text).center(((block_size + 1) * n_blocks) - 1, '-'))

	def print_line(values: Iterable[int], block_size: int = 3) -> None:
		values = list(values)
		mid = len(values) // 2

		colour: Optional[bool] = resolve_color_default()
		echo = partial(click.echo, nl=False, color=colour)

		for code in values[:mid]:
			echo(Colour.from_256_code(code, background=True)(str(code).center(block_size)) + ' ')
		with Fore.BLACK:
			for code in values[mid:-1]:
				echo(Colour.from_256_code(code, background=True)(str(code).center(block_size)) + ' ')
			click.echo(
					Colour.from_256_code(values[-1], background=True)(str(values[-1]).center(block_size).rstrip()),
					color=colour
					)

	print_heading(
			"Standard Colours".center((9 * 8) - 1, '-') + ' ' + "High-Intensity Colours".center((9 * 8) - 1, '-'),
			block_size=8,
			n_blocks=16
			)

	print_line(range(16), block_size=8)

	print_heading("216 Colours")
	for row in chunks(range(16, 232), 36):  # pylint: disable=loop-invariant-statement
		print_line(row)

	print_heading("Greyscale Colours", block_size=5, n_blocks=24)
	print_line(range(232, 256), block_size=5)

	click.echo('\n')


class AnsiCodes(ABC):
	"""
	Abstract base class for ANSI Codes.
	"""

	_stack: Union[Deque[str], List[str]]
	_reset: str

	def __init_subclass__(cls, **kwargs) -> None:
		"""
		The subclasses declare class attributes which are numbers.

		Upon instantiation we define instance attributes, which are the same
		as the class attributes but wrapped with the ANSI escape sequence.
		"""

		for name in dir(cls):
			if not name.startswith('_'):
				value = getattr(cls, name)
				setattr(cls, name, Colour(code_to_chars(value), cls._stack, cls._reset))

	def __new__(cls: Type[_AC], *args, **kwargs) -> Type[_AC]:  # noqa: D102
		return cls


class AnsiCursor:
	"""
	Provides methods to control the cursor.
	"""

	def UP(self, n: int = 1) -> str:
		"""
		Moves the cursor up ``n`` lines.

		:param n:
		"""

		return f"{CSI}{str(n)}A"

	def DOWN(self, n: int = 1) -> str:
		"""
		Moves the cursor down ``n`` lines.

		:param n:
		"""

		return f"{CSI}{str(n)}B"

	def FORWARD(self, n: int = 1) -> str:
		"""
		Moves the cursor forward (right) ``n`` lines.

		:param n:
		"""

		return f"{CSI}{str(n)}C"

	def BACK(self, n: int = 1) -> str:
		"""
		Moves the cursor backward (left) ``n`` lines.

		:param n:
		"""

		return f"{CSI}{str(n)}D"

	def POS(self, x: int = 1, y: int = 1) -> str:
		"""
		Moves the cursor to the given position.

		:param x:
		:param y:
		"""

		return f"{CSI}{str(y)};{str(x)}H"

	def HIDE(self) -> str:
		"""
		Hides the cursor.

		.. versionadded:: 0.7.0
		"""

		return f"{CSI}?25l"

	def SHOW(self) -> str:
		"""
		Shows the cursor.

		.. versionadded:: 0.7.0
		"""

		return f"{CSI}?25h"


class AnsiFore(AnsiCodes):
	r"""
	ANSI Colour Codes for foreground colour.

	The colours can be used as a context manager, a string, or a function.

	Valid values are:

	.. raw:: latex

		\vspace{-5px}
		\begin{multicols}{3}

	* BLACK
	* RED
	* GREEN
	* YELLOW
	* BLUE
	* MAGENTA
	* CYAN
	* WHITE
	* RESET
	* LIGHTBLACK_EX
	* LIGHTRED_EX
	* LIGHTGREEN_EX
	* LIGHTYELLOW_EX
	* LIGHTBLUE_EX
	* LIGHTMAGENTA_EX
	* LIGHTCYAN_EX
	* LIGHTWHITE_EX

	.. raw:: latex

		\end{multicols}
		\vspace{-5px}

	This class is also available under the shorter alias :py:obj:`Fore`.
	"""

	_stack = fore_stack
	_reset = f"{CSI}39m"

	BLACK = 30
	RED = 31
	GREEN = 32
	YELLOW = 33
	BLUE = 34
	MAGENTA = 35
	CYAN = 36
	WHITE = 37
	RESET = 39

	# These are fairly well supported, but not part of the standard.
	LIGHTBLACK_EX = 90
	LIGHTRED_EX = 91
	LIGHTGREEN_EX = 92
	LIGHTYELLOW_EX = 93
	LIGHTBLUE_EX = 94
	LIGHTMAGENTA_EX = 95
	LIGHTCYAN_EX = 96
	LIGHTWHITE_EX = 97


class AnsiBack(AnsiCodes):
	r"""
	ANSI Colour Codes for background colour.

	The colours can be used as a context manager, a string, or a function.

	Valid values are:

	.. raw:: latex

		\vspace{-5px}
		\begin{multicols}{3}

	* BLACK
	* RED
	* GREEN
	* YELLOW
	* BLUE
	* MAGENTA
	* CYAN
	* WHITE
	* RESET
	* LIGHTBLACK_EX
	* LIGHTRED_EX
	* LIGHTGREEN_EX
	* LIGHTYELLOW_EX
	* LIGHTBLUE_EX
	* LIGHTMAGENTA_EX
	* LIGHTCYAN_EX
	* LIGHTWHITE_EX

	.. raw:: latex

		\end{multicols}
		\vspace{-5px}

	This class is also available under the shorter alias :py:obj:`Back`.
	"""

	_stack = back_stack
	_reset = f"{CSI}49m"

	BLACK = 40
	RED = 41
	GREEN = 42
	YELLOW = 43
	BLUE = 44
	MAGENTA = 45
	CYAN = 46
	WHITE = 47
	RESET = 49

	# These are fairly well supported, but not part of the standard.
	LIGHTBLACK_EX = 100
	LIGHTRED_EX = 101
	LIGHTGREEN_EX = 102
	LIGHTYELLOW_EX = 103
	LIGHTBLUE_EX = 104
	LIGHTMAGENTA_EX = 105
	LIGHTCYAN_EX = 106
	LIGHTWHITE_EX = 107


class AnsiStyle(AnsiCodes):
	"""
	ANSI Colour Codes for text style.

	Valid values are:

	* BRIGHT
	* DIM
	* NORMAL

	Additionally, ``AnsiStyle.RESET_ALL`` can be used to reset the
	foreground and background colours as well as the text style.

	This class is also available under the shorter alias :py:obj:`Style`.
	"""

	_stack = style_stack
	_reset = f"{CSI}22m"

	BRIGHT = 1
	DIM = 2
	NORMAL = 22
	RESET_ALL = 0


Fore = AnsiFore
Back = AnsiBack
Style = AnsiStyle
Cursor = AnsiCursor()

fore_stack.append(Fore.RESET)
back_stack.append(Back.RESET)
style_stack.append(Style.NORMAL)

if __name__ == "__main__":
	print_256_colour_testpattern()



# stdlib
from typing import Optional

pytest_plugins = ("coincidence", "consolekit.testing")

try:
	# 3rd party
	from pytest_mypy_plugins.collect import YamlTestFile, pytest_addoption  # noqa: F401

	def pytest_collect_file(file_path, parent) -> Optional[YamlTestFile]:  # noqa: MAN001
		if file_path.suffix == ".yaml" and file_path.name.startswith(("test-", "test_")):
			return YamlTestFile.from_parent(parent, path=file_path, fspath=None)
		return None

except ImportError:
	pass



# stdlib
import inspect
from textwrap import dedent
from typing import Tuple

# 3rd party
import click
import pytest
from coincidence.regressions import AdvancedFileRegressionFixture
from coincidence.selectors import not_windows

# this package
import consolekit.commands
from consolekit import CONTEXT_SETTINGS, click_command, click_group
from consolekit.options import colour_option
from consolekit.terminal_colours import ColourTrilean
from consolekit.testing import CliRunner


@pytest.fixture()
def force_not_pycharm(monkeypatch) -> None:
	# Pretend we aren't running in PyCharm, even if we are.
	monkeypatch.setenv("PYCHARM_HOSTED", '0')
	monkeypatch.setattr(consolekit.utils, "_pycharm_terminal", lambda: False)


@pytest.fixture()
def markdown_demo_command() -> click.Command:

	@click_command(cls=consolekit.commands.MarkdownHelpCommand)
	def demo() -> None:
		"""
		This is the summary line.

		This **program** does *the* ``following``:

		* This
		* That
		* ~~The other~~ (deprecated)

		A
		multiline
		paragraph
		"""

	return demo


@pytest.fixture()
def markdown_demo_command_numbered() -> click.Command:

	@colour_option()
	@click_command(cls=consolekit.commands.MarkdownHelpCommand, no_args_is_help=True)
	def demo(colour: ColourTrilean = None) -> None:
		"""
		This is the summary line.

		This **program** does *the* ``following``:

		1. This
		2. That
		3. ~~The other~~ (deprecated)
		"""

	return demo


@pytest.fixture()
def markdown_demo_group() -> Tuple[click.Command, click.Command]:

	@click_group(cls=consolekit.commands.MarkdownHelpGroup)
	def demo() -> None:
		"""
		This is the summary line.

		This **program** does *the* ``following``:

		* This
		* That
		* ~~The other~~ (deprecated)
		"""

	@demo.command(cls=consolekit.commands.MarkdownHelpCommand)
	def foo() -> None:
		"""
		This is the summary line.

		This **program** does *the* ``following``:

		* This
		* That
		* ~~The other~~ (deprecated)
		"""

	return demo, foo


def test_raw_help_command(
		advanced_file_regression: AdvancedFileRegressionFixture,
		cli_runner: CliRunner,
		):

	@click_command(cls=consolekit.commands.RawHelpCommand)
	def demo() -> None:
		"""
		This is the summary line.

		This **program** does *the* ``following``:

		* This
		* That
		* ~~The other~~ (deprecated)
		"""

	@click_command(cls=consolekit.commands.RawHelpCommand)
	def helpless_demo() -> None:
		pass

	result = cli_runner.invoke(demo, args=["--help"])
	result.check_stdout(advanced_file_regression, extension=".md")

	assert demo.callback.__doc__ is not None
	expected = inspect.cleandoc(demo.callback.__doc__)
	assert dedent('\n'.join(result.stdout.splitlines()[2:-3])) == expected

	result = cli_runner.invoke(helpless_demo, args=["--help"])
	result.check_stdout(advanced_file_regression, extension=".helpless.md")

	assert helpless_demo.callback.__doc__ is None
	assert dedent('\n'.join(result.stdout.splitlines()[2:-3])) == ''


def test_raw_help_group(
		advanced_file_regression: AdvancedFileRegressionFixture,
		cli_runner: CliRunner,
		):

	@click_group(cls=consolekit.commands.RawHelpGroup)
	def demo() -> None:
		"""
		This is the summary line.

		This **program** does *the* ``following``:

		* This
		* That
		* ~~The other~~ (deprecated)
		"""

	@demo.command(cls=consolekit.commands.RawHelpCommand)
	def foo() -> None:
		"""
		This is the summary line.

		This **program** does *the* ``following``:

		* This
		* That
		* ~~The other~~ (deprecated)
		"""

	result = cli_runner.invoke(demo, args=["--help"])
	result.check_stdout(advanced_file_regression, extension="_group.md")

	result = cli_runner.invoke(foo, args=["--help"])
	result.check_stdout(advanced_file_regression, extension="_command.md")


@pytest.mark.usefixtures("force_not_pycharm")
@not_windows("Windows support for bold and italics is non-existent.")
def test_markdown_help_command(
		advanced_file_regression: AdvancedFileRegressionFixture,
		markdown_demo_command: click.Command,
		cli_runner: CliRunner,
		):
	result = cli_runner.invoke(markdown_demo_command, args=["--help"], color=True)
	result.check_stdout(advanced_file_regression, extension=".md")


@pytest.mark.usefixtures("force_not_pycharm")
def test_markdown_help_command_no_help(
		advanced_file_regression: AdvancedFileRegressionFixture,
		cli_runner: CliRunner,
		):

	@click_command(cls=consolekit.commands.MarkdownHelpCommand)
	def helpless_demo() -> None:
		pass

	result = cli_runner.invoke(helpless_demo, args=["--help"], color=True)
	result.check_stdout(advanced_file_regression, extension=".md")


@pytest.mark.usefixtures("force_not_pycharm")
@not_windows("Windows support for bold and italics is non-existent.")
def test_markdown_help_group(
		advanced_file_regression: AdvancedFileRegressionFixture,
		markdown_demo_group: Tuple[click.Command, click.Command],
		cli_runner: CliRunner,
		):

	demo_group, demo_command = markdown_demo_group

	result = cli_runner.invoke(demo_group, args=["--help"], color=True)
	result.check_stdout(advanced_file_regression, extension="_group.md")

	result = cli_runner.invoke(demo_command, args=["--help"], color=True)
	result.check_stdout(advanced_file_regression, extension="_command.md")


@pytest.mark.usefixtures("force_not_pycharm")
@not_windows("Windows support for bold and italics is non-existent.")
def test_markdown_help_command_ordered_list(
		advanced_file_regression: AdvancedFileRegressionFixture,
		cli_runner: CliRunner,
		):

	@click_command(cls=consolekit.commands.MarkdownHelpCommand)
	def demo() -> None:
		"""
		This is the summary line.

		This **program** does *the* ``following``:

		1. This
		2. That
		3. ~~The other~~ (deprecated)
		"""

	result = cli_runner.invoke(demo, args=["--help"], color=True)
	result.check_stdout(advanced_file_regression, extension=".md")


@not_windows("Windows support for bold and italics is non-existent.")
def test_markdown_help_command_pycharm(
		advanced_file_regression: AdvancedFileRegressionFixture,
		monkeypatch,
		markdown_demo_command: click.Command,
		cli_runner: CliRunner,
		):
	monkeypatch.setenv("PYCHARM_HOSTED", '1')
	monkeypatch.setattr(consolekit.utils, "_pycharm_terminal", lambda: False)

	result = cli_runner.invoke(markdown_demo_command, args=["--help"], color=True)
	result.check_stdout(advanced_file_regression, extension=".md")


def test_private_helper(monkeypatch):
	monkeypatch.setenv("PYCHARM_HOSTED", '1')

	assert not consolekit.utils._pycharm_terminal()


@pytest.mark.usefixtures("force_not_pycharm")
def test_markdown_help_command_no_colour(
		advanced_file_regression: AdvancedFileRegressionFixture,
		markdown_demo_command_numbered: click.Command,
		cli_runner: CliRunner,
		monkeypatch,
		):

	result = cli_runner.invoke(markdown_demo_command_numbered, args=["--help", "--no-colour"], color=True)
	result.check_stdout(advanced_file_regression, extension=".md")

	# Again with envvar
	monkeypatch.setenv("NO_COLOR", '1')
	result = cli_runner.invoke(markdown_demo_command_numbered, args=["--help"], color=True)
	result.check_stdout(advanced_file_regression, extension=".md")


@pytest.mark.usefixtures("force_not_pycharm")
def test_markdown_help_no_args_is_help(
		advanced_file_regression: AdvancedFileRegressionFixture,
		markdown_demo_command_numbered: click.Command,
		cli_runner: CliRunner,
		):
	result = cli_runner.invoke(markdown_demo_command_numbered)
	result.check_stdout(advanced_file_regression, extension=".md")
	assert result.exit_code == 0


@pytest.mark.usefixtures("force_not_pycharm")
def test_markdown_help_group_no_args_is_help(
		advanced_file_regression: AdvancedFileRegressionFixture,
		markdown_demo_group: Tuple[click.Command, click.Command],
		cli_runner: CliRunner,
		):
	result = cli_runner.invoke(markdown_demo_group[0])
	result.check_stdout(advanced_file_regression, extension=".md")
	assert result.exit_code == 0


def test_suggestion_group(
		advanced_file_regression: AdvancedFileRegressionFixture,
		cli_runner: CliRunner,
		):

	@click_group(context_settings={**CONTEXT_SETTINGS, "token_normalize_func": lambda x: x.lower()})
	def demo() -> None:
		"""
		A program.
		"""

	@demo.command()  # skipcq
	def search() -> None:
		"""
		Conduct a search.
		"""

	result = cli_runner.invoke(demo, args=["searh"])
	result.check_stdout(advanced_file_regression, extension="_success.md")
	assert result.exit_code == 2

	result = cli_runner.invoke(demo, args=["list"])
	result.check_stdout(advanced_file_regression, extension="_failure.md")
	assert result.exit_code == 2

	result = cli_runner.invoke(demo, args=["SEARCH"])
	assert not result.stdout.rstrip()
	assert result.exit_code == 0


def test_context_inheriting_group(
		advanced_file_regression: AdvancedFileRegressionFixture,
		cli_runner: CliRunner,
		):

	@click_group(cls=consolekit.commands.ContextInheritingGroup)
	def demo() -> None:
		"""
		A program.
		"""

	@demo.command()
	def search() -> None:
		"""
		Conduct a search.
		"""

	@demo.group()
	def foo() -> None:
		"""
		Does bar.
		"""

	result = cli_runner.invoke(demo, args=["--help"])
	assert result.exit_code == 0

	result = cli_runner.invoke(demo, args=["-h"])
	assert result.exit_code == 0
	result.check_stdout(advanced_file_regression, extension=".md")

	result = cli_runner.invoke(foo, args=["--help"])
	assert result.exit_code == 0

	result = cli_runner.invoke(foo, args=["-h"])
	assert result.exit_code == 0
	result.check_stdout(advanced_file_regression, extension="_foo.md")

	result = cli_runner.invoke(search, args=["--help"])
	assert result.exit_code == 0

	result = cli_runner.invoke(search, args=["-h"])
	assert result.exit_code == 0
	result.check_stdout(advanced_file_regression, extension="_search.md")



# stdlib
import math
import os
import shutil
import sys
from typing import Tuple

# 3rd party
import click
from coincidence.regressions import AdvancedDataRegressionFixture, check_file_regression
from coincidence.selectors import not_windows
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.utils import redirect_output
from pytest_regressions.file_regression import FileRegressionFixture

# this package
import consolekit
from consolekit import click_command
from consolekit.terminal_colours import ColourTrilean
from consolekit.utils import (
		abort,
		coloured_diff,
		hidden_cursor,
		hide_cursor,
		import_commands,
		is_command,
		long_echo,
		overtype,
		show_cursor,
		solidus_spinner
		)


def test_overtype(capsys):
	print("Waiting...", end='')
	overtype("foo", "bar")
	sys.stdout.flush()

	captured = capsys.readouterr()
	stdout = captured.out.split('\n')
	assert stdout == ["Waiting...\rfoo bar"]

	print("Waiting...", end='')
	overtype("foo", "bar", sep='')
	sys.stdout.flush()

	captured = capsys.readouterr()
	stdout = captured.out.split('\n')
	assert stdout == ["Waiting...\rfoobar"]

	print("Waiting...", end='')
	overtype("foo", "bar", sep='-', end='\n')
	sys.stdout.flush()

	captured = capsys.readouterr()
	stdout = captured.out.split('\n')
	assert stdout == ["Waiting...\rfoo-bar", '']

	sys.stderr.write("Waiting...")
	overtype("foo", "bar", file=sys.stderr)
	sys.stdout.flush()

	captured = capsys.readouterr()
	stderr = captured.err.split('\n')
	assert stderr == ["Waiting...\rfoo bar"]


def test_coloured_diff(file_regression: FileRegressionFixture):
	data_dir = PathPlus(__file__).parent / "test_diff_"
	original = data_dir / "original"
	modified = data_dir / "modified"

	diff = coloured_diff(
			original.read_lines(),
			modified.read_lines(),
			fromfile="original_file.txt",
			tofile="modified_file.txt",
			fromfiledate="(original)",
			tofiledate="(modified)",
			lineterm='',
			)

	check_file_regression(diff, file_regression)


def test_is_command():

	@click_command()
	def main() -> None: ...

	assert is_command(main)
	assert not is_command(int)
	assert not is_command(lambda: True)
	assert not is_command(math.ceil)


def test_hidden_cursor(monkeypatch, capsys, advanced_data_regression: AdvancedDataRegressionFixture):
	monkeypatch.setattr(consolekit.terminal_colours, "resolve_color_default", lambda *args: True)

	hide_cursor()
	show_cursor()

	with hidden_cursor():
		click.echo(f"\r{next(solidus_spinner)}", nl=False)
		click.echo(f"\r{next(solidus_spinner)}", nl=False)
		click.echo(f"\r{next(solidus_spinner)}", nl=False)

	advanced_data_regression.check(tuple(capsys.readouterr()))


def test_import_commands():
	# this package
	from tests import import_commands_demo

	commands = import_commands(import_commands_demo)
	assert commands == [
			import_commands_demo.command1,
			import_commands_demo.commando,
			import_commands_demo.submodule.command2,
			import_commands_demo.submodule.group2,
			]


def test_long_echo(monkeypatch):

	def get_terminal_size(fallback: Tuple[int, int] = (80, 24)) -> Tuple[int, int]:
		return os.terminal_size((80, 5))

	def echo_via_pager(text_or_generator, color: ColourTrilean = None) -> None:  # noqa: MAN001
		click.echo('\n'.join(f"|{line}" for line in text_or_generator.splitlines()))

	monkeypatch.setattr(shutil, "get_terminal_size", get_terminal_size)
	monkeypatch.setattr(click, "echo_via_pager", echo_via_pager)

	with redirect_output() as (stdout, stderr):
		stdout.isatty = lambda *args: True  # type: ignore[assignment]
		assert stdout.isatty()
		assert sys.stdout.isatty()

		long_echo([
				"Line 1",
				"Line 2",
				"Line 3",
				"Line 4",
				"Line 5",
				"Line 6",
				])

	assert stdout.getvalue() == "|Line 1\n|Line 2\n|Line 3\n|Line 4\n|Line 5\n|Line 6\n"

	with redirect_output() as (stdout, stderr):
		stdout.isatty = lambda *args: True  # type: ignore[assignment]
		assert stdout.isatty()
		assert sys.stdout.isatty()

		long_echo('\n'.join([
				"Line 1",
				"Line 2",
				"Line 3",
				"Line 4",
				"Line 5",
				]))

	assert stdout.getvalue() == "Line 1\nLine 2\nLine 3\nLine 4\nLine 5\n"


@not_windows(reason="Output differs on Windows")
def test_abort(capsys):
	abort("The program will now abort.", colour=True)
	assert capsys.readouterr().err == "\x1b[31mThe program will now abort.\x1b[39m\n"


def test_abort_no_colour(capsys):
	abort("The program will now abort.", colour=False)
	assert capsys.readouterr().err == "The program will now abort.\n"



# stdlib
import sys
from typing import Iterable

# 3rd party
import click
from pytest_regressions.file_regression import FileRegressionFixture

# this package
from consolekit import click_command
from consolekit.options import (
		DescribedArgument,
		MultiValueOption,
		auto_default_argument,
		auto_default_option,
		colour_option,
		flag_option,
		force_option,
		no_pager_option,
		verbose_option,
		version_option
		)
from consolekit.terminal_colours import ColourTrilean
from consolekit.testing import CliRunner, Result


def test_described_argument(
		file_regression: FileRegressionFixture,
		cli_runner: CliRunner,
		):

	@click.argument(
			"dest",
			cls=DescribedArgument,
			type=click.STRING,
			description="The destination directory.",
			)
	@click_command()
	def main(dest: str) -> None:
		print(dest)

	result = cli_runner.invoke(main, args="--help")
	result.check_stdout(file_regression, extension=".md")
	assert result.exit_code == 0

	result = cli_runner.invoke(main, catch_exceptions=False, args="./staging")
	assert result.stdout.rstrip() == "./staging"
	assert result.exit_code == 0

	ctx = click.Context(main, info_name="main", parent=None)
	argument = ctx.command.params[0]
	assert isinstance(argument, DescribedArgument)
	assert argument.description == "The destination directory."


def test_auto_default_option(
		file_regression: FileRegressionFixture,
		cli_runner: CliRunner,
		):

	@auto_default_option(
			"--width",
			type=click.INT,
			help="The max width to display.",
			show_default=True,
			)
	@click_command()
	def main(width: int = 80) -> None:
		print(width)

	argument = main.params[0]
	assert argument.default == 80
	assert not argument.required
	assert argument.show_default  # type: ignore[attr-defined]

	result = cli_runner.invoke(main, args="--help")
	result.check_stdout(file_regression, extension=".md")
	assert result.exit_code == 0

	result = cli_runner.invoke(main, catch_exceptions=False)
	assert result.stdout.rstrip() == "80"
	assert result.exit_code == 0

	for width in [0, 8, 36, 48, 130]:

		result = cli_runner.invoke(main, args=["--width", str(width)])
		assert result.stdout.rstrip() == str(width)
		assert result.exit_code == 0


def test_flag_option(cli_runner: CliRunner):

	@flag_option("--no-colour")
	@click_command()
	def main(no_colour: bool) -> None:
		print(no_colour)

	result = cli_runner.invoke(main, catch_exceptions=False)
	assert result.stdout.rstrip() == "False"
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args="--no-colour")
	assert result.stdout.rstrip() == "True"
	assert result.exit_code == 0


def test_no_pager_option(file_regression: FileRegressionFixture, cli_runner: CliRunner):

	@no_pager_option()
	@click_command()
	def main(no_pager: bool) -> None:
		print(no_pager)

	result = cli_runner.invoke(main, catch_exceptions=False)
	assert result.stdout.rstrip() == "False"
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args="--no-pager")
	assert result.stdout.rstrip() == "True"
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args="-h")
	result.check_stdout(file_regression)
	assert result.exit_code == 0


def test_force_option(file_regression: FileRegressionFixture, cli_runner: CliRunner):

	@force_option("Force the operation")
	@click_command()
	def main(force: bool) -> None:
		print(force)

	result = cli_runner.invoke(main, catch_exceptions=False)
	assert result.stdout.rstrip() == "False"
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args="--force")
	assert result.stdout.rstrip() == "True"
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args="-f")
	assert result.stdout.rstrip() == "True"
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args="-h")
	result.check_stdout(file_regression)
	assert result.exit_code == 0


def test_verbose_option(file_regression: FileRegressionFixture, cli_runner: CliRunner):

	@verbose_option("Show verbose output.")
	@click_command()
	def main(verbose: int) -> None:
		print(verbose)

	result = cli_runner.invoke(main)
	assert result.stdout.rstrip() == '0'
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args="--verbose")
	assert result.stdout.rstrip() == '1'
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args="-v")
	assert result.stdout.rstrip() == '1'
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args=["--verbose", "--verbose"])
	assert result.stdout.rstrip() == '2'
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args="-vv")
	assert result.stdout.rstrip() == '2'
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args="-vvv")
	assert result.stdout.rstrip() == '3'
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args="-h")
	result.check_stdout(file_regression)
	assert result.exit_code == 0


def test_colour_option(file_regression: FileRegressionFixture, cli_runner: CliRunner):

	@colour_option()
	@click_command()
	def main(colour: ColourTrilean) -> None:
		print(colour)

	result = cli_runner.invoke(main, catch_exceptions=False)
	assert result.stdout.rstrip() == "None"
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args="--no-colour")
	assert result.stdout.rstrip() == "False"
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args="--colour")
	assert result.stdout.rstrip() == "True"
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args="-h")
	result.check_stdout(file_regression)
	assert result.exit_code == 0


def test_version_option(file_regression: FileRegressionFixture, cli_runner: CliRunner):

	def version_callback(ctx: click.Context, param: click.Option, value: int) -> None:
		if not value or ctx.resilient_parsing:
			return

		if value > 1:
			click.echo("consolekit version 1.2.3, Python 3.8.5")
		else:
			click.echo("consolekit version 1.2.3")

		ctx.exit()

	@version_option(version_callback)
	@click_command()
	def main() -> None:
		sys.exit(1)

	result = cli_runner.invoke(main, args="--version")
	assert result.stdout.rstrip() == "consolekit version 1.2.3"
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args=["--version", "--version"])
	assert result.stdout.rstrip() == "consolekit version 1.2.3, Python 3.8.5"
	assert result.exit_code == 0


def test_multi_value_option(cli_runner: CliRunner):

	@click.option(
			"--select",
			type=click.STRING,
			help="The checks to enable",
			cls=MultiValueOption,
			)
	@colour_option()
	@click_command()
	def main(select: Iterable[str], colour: bool) -> None:
		select = list(select)
		print(*select)
		print(", ".join(select))

	result: Result = cli_runner.invoke(main, catch_exceptions=False)
	assert result.stdout.rstrip() == ''
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args=["--select", "E102", "F223"])
	assert result.stdout.rstrip() == "E102 F223\nE102, F223"
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args=["--select", "E102", "F223", "--colour"])
	assert result.stdout.rstrip() == "E102 F223\nE102, F223"
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args=["--select", "E102"])
	assert result.stdout.rstrip() == "E102\nE102"
	assert result.exit_code == 0


def test_multi_value_option_int(cli_runner: CliRunner):

	@click.option("--value", help="Some values", cls=MultiValueOption, type=int, default=(1, 2, 3))
	@colour_option()
	@click_command()
	def main(value: Iterable[int], colour: bool) -> None:
		print(value)

	result: Result = cli_runner.invoke(main, catch_exceptions=False)
	assert result.stdout.rstrip() == "(1, 2, 3)"
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args=["--value", '4', '5', '6'])
	assert result.stdout.rstrip() == "(4, 5, 6)"
	assert result.exit_code == 0


def test_auto_default_argument(cli_runner: CliRunner):

	@auto_default_argument(
			"greeting",
			type=click.STRING,
			)
	@click_command()
	def main(greeting: str = "Hello") -> None:
		print(f"{greeting} User!")

	argument = main.params[0]
	assert argument.default == "Hello"
	assert not argument.required

	result = cli_runner.invoke(main)
	assert result.stdout.rstrip() == "Hello User!"
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args=["Good Morning"])
	assert result.stdout.rstrip() == "Good Morning User!"
	assert result.exit_code == 0



# Demo of consolekit's spinners

# stdlib
import time

# this package
from consolekit.utils import braille_spinner, hidden_cursor, snake_spinner, solidus_spinner

REPEATS = 50
SLEEP = 0.5

with hidden_cursor():

	for _ in range(REPEATS):
		print(f"\r{next(braille_spinner)}", end='')
		time.sleep(SLEEP)

	for _ in range(REPEATS):
		print(f"\r{next(solidus_spinner)}", end='')
		time.sleep(SLEEP)

	for _ in range(REPEATS):
		print(f"\r{next(snake_spinner)}", end='')
		time.sleep(SLEEP)



# this package
from consolekit import click_command, click_group

# this package
from . import submodule  # noqa: F401  # skipcq


@click_group()
def commando():
	pass


@click_command()
def command1():
	pass



# 3rd party
from domdf_python_tools.stringlist import StringList
from sphinx import addnodes
from sphinx.application import Sphinx  # nodep
from sphinx.config import Config  # nodep
from sphinx.writers.latex import LaTeXTranslator
from sphinxcontrib import toctree_plus


def visit_desc(translator: LaTeXTranslator, node: addnodes.desc) -> None:
	"""
	Visit an :class:`addnodes.desc` node and add a custom table of contents label for the item, if required.

	.. versionadded:: 0.3.0

	:param translator:
	:param node:
	"""

	translator.body.append(r"\needspace{5\baselineskip}")
	toctree_plus.visit_desc(translator, node)


def configure(app: Sphinx, config: Config):
	"""
	Configure Sphinx Extension.

	:param app: The Sphinx application.
	:param config:
	"""

	latex_elements = getattr(config, "latex_elements", {})

	latex_extrapackages = StringList(latex_elements.get("extrapackages", ''))
	latex_extrapackages.append(r"\usepackage{needspace}")
	latex_elements["extrapackages"] = str(latex_extrapackages)

	config.latex_elements = latex_elements


def setup(app: Sphinx):
	"""
	Setup Sphinx Extension.

	:param app: The Sphinx application.
	"""

	app.connect("config-inited", configure)
	app.add_node(addnodes.desc, latex=(visit_desc, toctree_plus.depart_desc), override=True)

	return {"parallel_read_safe": True}



# stdlib
import sys

# this package
from consolekit import click_command
from consolekit.options import version_option
from consolekit.testing import CliRunner
from consolekit.versions import get_formatted_versions, get_version_callback


def test_get_formatted_versions():
	dependencies = ["click", "deprecation-alias", "domdf-python-tools", "mistletoe", "typing-extensions"]

	sl = get_formatted_versions(dependencies)
	for line, dep in zip(sl, dependencies):
		assert line.startswith(f"{dep}:")
	assert sl[-2].startswith("Python: 3.")

	sl = get_formatted_versions(dependencies, show_platform=False)
	for line, dep in zip(sl, [*dependencies]):
		assert line.startswith(f"{dep}:")
	assert sl[-1].startswith("Python: 3.")

	sl = get_formatted_versions(dependencies, show_python=False)
	assert len(sl) == len(dependencies) + 1

	sl = get_formatted_versions(dependencies, show_python=False, show_platform=False)
	assert len(sl) == len(dependencies)

	sl = get_formatted_versions({
			"click": "pkg1",
			"deprecation-alias": "pkg2",
			"domdf-python-tools": "pkg3",
			"mistletoe": "pkg4",
			"typing-extensions": "pkg5"
			})
	for line, name in zip(sl, ["pkg1", "pkg2", "pkg3", "pkg4", "pkg5"]):
		assert line.startswith(f"{name}:")
	assert sl[-2].startswith("Python: 3.")


def test_version_callback(cli_runner: CliRunner):

	@version_option(
			get_version_callback(
					"1.2.3",
					"my-tool",
					["click", "deprecation-alias", "domdf-python-tools", "mistletoe", "typing-extensions"]
					)
			)
	@click_command()
	def main() -> None:
		sys.exit(1)

	result = cli_runner.invoke(main, args="--version")
	assert result.stdout.rstrip() == "my-tool version 1.2.3"
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args=["--version", "--version"])
	assert result.stdout.startswith("my-tool version 1.2.3, Python 3.")
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args=["--version", "--version", "--version"])
	print(result.stdout)
	assert result.stdout.startswith("my-tool\n  Version: 1.2.3\n  click: ")
	assert result.exit_code == 0



#!/usr/bin/env python3
#
#  _types.py
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import Any, Callable, Optional, Tuple, Union

# 3rd party
from click import Context, Parameter, ParamType

_ConvertibleType = Union[
	type,
	ParamType,
	Tuple[Union[type, ParamType], ...],
	Callable[[str], Any],
	Callable[[Optional[str]], Any]
	]

Callback = Callable[[Context, Parameter, str], Any]



# From https://github.com/tartley/colorama
# Copyright Jonathan Hartley 2013
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holders, nor those of its contributors
#   may be used to endorse or promote products derived from this software without
#   specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# stdlib
import sys

# 3rd party
import pytest

# this package
from consolekit.terminal_colours import Back, Colour, Fore, Style

stdout_orig = sys.stdout
stderr_orig = sys.stderr

colorama = pytest.importorskip("colorama")
AnsiToWin32 = colorama.ansitowin32.AnsiToWin32


def setup_module() -> None:
	# sanity check: stdout should be a file or StringIO object.
	# It will only be AnsiToWin32 if init() has previously wrapped it
	assert not isinstance(sys.stdout, AnsiToWin32)
	assert not isinstance(sys.stderr, AnsiToWin32)


def teardown_module() -> None:
	sys.stdout = stdout_orig
	sys.stderr = stderr_orig


@pytest.mark.parametrize(
		"obj, expects",
		[
				(Fore.BLACK, "\u001b[30m"),
				(Fore.RED, "\u001b[31m"),
				(Fore.GREEN, "\u001b[32m"),
				(Fore.YELLOW, "\u001b[33m"),
				(Fore.BLUE, "\u001b[34m"),
				(Fore.MAGENTA, "\u001b[35m"),
				(Fore.CYAN, "\u001b[36m"),
				(Fore.WHITE, "\u001b[37m"),
				(Fore.RESET, "\u001b[39m"),

				# Check the light, extended versions.
				(Fore.LIGHTBLACK_EX, "\u001b[90m"),
				(Fore.LIGHTRED_EX, "\u001b[91m"),
				(Fore.LIGHTGREEN_EX, "\u001b[92m"),
				(Fore.LIGHTYELLOW_EX, "\u001b[93m"),
				(Fore.LIGHTBLUE_EX, "\u001b[94m"),
				(Fore.LIGHTMAGENTA_EX, "\u001b[95m"),
				(Fore.LIGHTCYAN_EX, "\u001b[96m"),
				(Fore.LIGHTWHITE_EX, "\u001b[97m"),
				],
		)
def test_fore_attributes(obj: Colour, expects: str, capsys):
	assert obj == expects
	assert obj("Hello World") == f"{obj}Hello World\033[39m"

	with obj:
		print("Hello World!")
		with Fore.RESET:
			print("Reset!")
		print("Coloured Again!")
	print("Reset Again!")

	captured = capsys.readouterr()
	stdout = captured.out.split('\n')
	assert stdout[0] == f"{obj}Hello World!"
	assert stdout[1] == f"\033[39mReset!"
	assert stdout[2] == f"{obj}Coloured Again!"
	assert stdout[3] == f"\033[39mReset Again!"
	assert stdout[4] == ''


@pytest.mark.parametrize(
		"obj, expects",
		[
				(Back.BLACK, "\u001b[40m"),
				(Back.RED, "\u001b[41m"),
				(Back.GREEN, "\u001b[42m"),
				(Back.YELLOW, "\u001b[43m"),
				(Back.BLUE, "\u001b[44m"),
				(Back.MAGENTA, "\u001b[45m"),
				(Back.CYAN, "\u001b[46m"),
				(Back.WHITE, "\u001b[47m"),
				(Back.RESET, "\u001b[49m"),

				# Check the light, extended versions.
				(Back.LIGHTBLACK_EX, "\u001b[100m"),
				(Back.LIGHTRED_EX, "\u001b[101m"),
				(Back.LIGHTGREEN_EX, "\u001b[102m"),
				(Back.LIGHTYELLOW_EX, "\u001b[103m"),
				(Back.LIGHTBLUE_EX, "\u001b[104m"),
				(Back.LIGHTMAGENTA_EX, "\u001b[105m"),
				(Back.LIGHTCYAN_EX, "\u001b[106m"),
				(Back.LIGHTWHITE_EX, "\u001b[107m"),
				],
		)
def test_back_attributes(obj: Colour, expects: str, capsys):
	assert obj == expects
	assert obj("Hello World") == f"{obj}Hello World\033[49m"

	with obj:
		print("Hello World!")
		with Back.RESET:
			print("Reset!")
		print("Coloured Again!")
	print("Reset Again!")

	captured = capsys.readouterr()
	stdout = captured.out.split('\n')
	assert stdout[0] == f"{obj}Hello World!"
	assert stdout[1] == f"\033[49mReset!"
	assert stdout[2] == f"{obj}Coloured Again!"
	assert stdout[3] == f"\033[49mReset Again!"
	assert stdout[4] == ''


@pytest.mark.parametrize(
		"obj, expects",
		[
				(Style.DIM, "\u001b[2m"),
				(Style.NORMAL, "\u001b[22m"),
				(Style.BRIGHT, "\u001b[1m"),
				],
		)
def test_style_attributes(obj: Colour, expects: str, capsys):
	assert obj == expects
	assert obj("Hello World") == f"{obj}Hello World\033[22m"

	with obj:
		print("Hello World!")
		with Style.NORMAL:
			print("Reset!")
		print("Coloured Again!")
	print("Reset Again!")

	captured = capsys.readouterr()
	stdout = captured.out.split('\n')
	assert stdout[0] == f"{obj}Hello World!"
	assert stdout[1] == f"\033[22mReset!"
	assert stdout[2] == f"{obj}Coloured Again!"
	assert stdout[3] == f"\033[22mReset Again!"
	assert stdout[4] == ''



#!/usr/bin/env python3
#
#  versions.py
"""
Tool to get software versions.

.. versionadded:: 1.6.0
"""
#
#  Copyright © 2023 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import platform
import sys
import textwrap
from typing import Any, Callable, Iterable, Mapping, Union

# 3rd party
import click
from domdf_python_tools.compat import importlib_metadata
from domdf_python_tools.stringlist import StringList
from domdf_python_tools.words import LF

__all__ = ("get_formatted_versions", "get_version_callback")


def get_formatted_versions(
		dependencies: Union[Iterable[str], Mapping[str, str]] = (),
		show_python: bool = True,
		show_platform: bool = True,
		) -> StringList:
	"""
	Return the versions of software dependencies, one per line.

	:param dependencies: Either a list of dependency names,
	    or a mapping of dependency name to a more human-readable form.
	:param show_python:
	:param show_platform:
	"""

	versions = StringList()

	if not isinstance(dependencies, Mapping):
		dependencies = {d: d for d in dependencies}

	for dependency_name, display_name in dependencies.items():

		dep_version = importlib_metadata.version(dependency_name)
		versions.append(f"{display_name}: {dep_version}")

	if show_python:
		versions.append(f"Python: {sys.version.replace(LF, ' ')}")

	if show_platform:
		versions.append(' '.join(platform.system_alias(platform.system(), platform.release(), platform.version())))

	return versions


def get_version_callback(
		tool_version: str,
		tool_name: str,
		dependencies: Union[Iterable[str], Mapping[str, str]] = (),
		) -> Callable[[click.Context, click.Option, int], Any]:
	"""
	Creates a callback for :class:`~.version_option`.

	With each ``--version`` argument the callback displays the package version,
	then adds the python version, and finally adds dependency versions.


	:param tool_version: The version of the tool to show the version of.
	:param tool_name: The name of the tool to show the version of.
	:param dependencies: Either a list of dependency names,
	    or a mapping of dependency name to a more human-readable form.
	"""

	def version_callback(ctx: click.Context, param: click.Option, value: int) -> None:
		"""
		Callback for displaying the package version (and optionally the Python runtime).
		"""

		if not value or ctx.resilient_parsing:
			return

		if value > 2:
			versions = textwrap.indent(
					get_formatted_versions(dependencies),  # type: ignore[arg-type]
					"  ",
					)
			click.echo(tool_name)
			click.echo(f"  Version: {tool_version}")
			click.echo(versions.rstrip())
		elif value > 1:
			python_version = sys.version.replace('\n', ' ')
			click.echo(f"{tool_name} version {tool_version}, Python {python_version}")
		else:
			click.echo(f"{tool_name} version {tool_version}")

		ctx.exit()

	return version_callback



#!/usr/bin/env python3
#
#  utils.py
"""
Utility functions.

.. versionchanged:: 1.0.0

	:func:`~.tracebacks.traceback_handler` and :func:`~.tracebacks.handle_tracebacks`
	moved to :mod:`consolekit.tracebacks`.
	They will still be importable from here until v2.0.0

.. automodulesumm:: consolekit.utils
.. latex:clearpage::
"""
#
#  Copyright © 2020-2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import contextlib
import difflib
import os
import shutil
import sys
from functools import lru_cache
from itertools import cycle
from types import ModuleType
from typing import IO, Any, Iterable, Iterator, List, Optional, Sequence, Union

# 3rd party
import click
import deprecation_alias
from domdf_python_tools.stringlist import StringList
from domdf_python_tools.words import SANS_SERIF_ITALIC_LETTERS
from mistletoe import block_token, span_token  # type: ignore[import]
from mistletoe.base_renderer import BaseRenderer  # type: ignore[import]
from typing_extensions import TypeGuard

# this package
from consolekit import terminal_colours, tracebacks
from consolekit.terminal_colours import ColourTrilean, resolve_color_default

__all__ = (
		"get_env_vars",
		"is_command",
		"import_commands",
		"abort",
		"overtype",
		"coloured_diff",
		"solidus_spinner",
		"braille_spinner",
		"snake_spinner",
		"hide_cursor",
		"show_cursor",
		"TerminalRenderer",
		"hidden_cursor",
		"long_echo",
		)

_deprecator = deprecation_alias.deprecated(
		deprecated_in="1.0.0",
		removed_in="2.0.0",
		current_version="1.7.2",
		details="Import from consolekit.tracebacks instead."
		)

handle_tracebacks = _deprecator(tracebacks.handle_tracebacks)
traceback_handler = _deprecator(tracebacks.traceback_handler)


def get_env_vars(ctx, args, incomplete) -> List[str]:  # noqa: D103,MAN001
	return [k for k in os.environ.keys() if incomplete in k]


def is_command(obj: Any) -> TypeGuard[click.Command]:
	"""
	Return whether ``obj`` is a click command.

	:param obj:
	"""

	return isinstance(obj, click.Command)


def import_commands(source: Optional[ModuleType] = None, entry_point: Optional[str] = None) -> List[click.Command]:
	"""
	Returns a list of all commands.

	Commands can be defined locally in the module given in ``source``,
	or by third party extensions who define an entry point in the following format:

	::

		<name (can be anything)> = <module name>:<command>

	:param source:
	:param entry_point:
	"""

	# 3rd party
	from domdf_python_tools.import_tools import discover, discover_entry_points

	all_commands = []

	if source is not None:
		all_commands.extend(discover(source, is_command, exclude_side_effects=False))
	if entry_point is not None:
		all_commands.extend(discover_entry_points(entry_point, is_command))

	return all_commands


# TODO: Turn this into a class so the message is only printed when raised.
def abort(message: str, colour: ColourTrilean = None) -> Exception:
	"""
	Aborts the program execution.

	:param message:
	:param colour: Whether to use coloured output. Default auto-detect.
	:no-default colour:

	.. versionchanged:: 1.0.1  Added the ``colour`` option.
	"""

	click.echo(terminal_colours.Fore.RED(message), err=True, color=resolve_color_default(colour))
	return click.Abort()


def overtype(
		*objects,
		sep: str = ' ',
		end: str = '',
		file: Optional[IO] = None,
		flush: bool = False,
		) -> None:
	r"""
	Print ``objects`` to the text stream ``file``, starting with ``"\r"``, separated by ``sep``
	and followed by ``end``.

	``sep``, ``end``, ``file`` and ``flush``, if present, must be given as keyword arguments

	All non-keyword arguments are converted to strings like :class:`str` does and written to the stream,
	separated by `sep` and followed by `end`.
	If no such arguments are given, :func:`~consolekit.utils.overtype` will just write ``"\r"``.

	.. TODO:: This does not currently work in the PyCharm console, at least on Windows

	:param objects: A list of strings or string-like objects to write to the terminal.
	:param sep: String to separate the objects with.
	:param end: String to end with.
	:param file: An object with a ``write(string)`` method.
	:default file: ``sys.stdout``
	:param flush: If :py:obj:`True`, the stream is forcibly flushed.
	"""  # noqa: D400

	object0 = f"\r{objects[0]}"
	objects = (object0, *objects[1:])
	print(*objects, sep=sep, end=end, file=file, flush=flush)


def coloured_diff(
		a: Sequence[str],
		b: Sequence[str],
		fromfile: str = '',
		tofile: str = '',
		fromfiledate: str = '',
		tofiledate: str = '',
		n: int = 3,
		lineterm: str = '\n',
		removed_colour: terminal_colours.Colour = terminal_colours.Fore.RED,
		added_colour: terminal_colours.Colour = terminal_colours.Fore.GREEN,
		) -> str:
	r"""
	Compare two sequences of lines; generate the delta as a unified diff.

	Unified diffs are a compact way of showing line changes and a few
	lines of context. The number of context lines is set by ``n`` which
	defaults to three.

	By default, the diff control lines (those with ``---``, ``+++``, or ``@@``)
	are created with a trailing newline. This is helpful so that inputs
	created from ``file.readlines()`` result in diffs that are suitable for
	``file.writelines()`` since both the inputs and outputs have trailing
	newlines.

	For inputs that do not have trailing newlines, set the lineterm
	argument to ``''`` so that the output will be uniformly newline free.

	The unidiff format normally has a header for filenames and modification
	times. Any or all of these may be specified using strings for
	``fromfile``, ``tofile``, ``fromfiledate``, and ``tofiledate``.
	The modification times are normally expressed in the ISO 8601 format.

	.. versionadded:: 0.3.0

	:bold-title:`Example:`

	>>> for line in coloured_diff(
	...     'one two three four'.split(),
	...     'zero one tree four'.split(), 'Original', 'Current',
	...     '2005-01-26 23:30:50', '2010-04-02 10:20:52',
	...     lineterm='',
	...     ):
	...     print(line)                 # doctest: +NORMALIZE_WHITESPACE
	--- Original        2005-01-26 23:30:50
	+++ Current         2010-04-02 10:20:52
	@@ -1,4 +1,4 @@
	+zero
	one
	-two
	-three
	+tree
	four

	:param a:
	:param b:
	:param fromfile:
	:param tofile:
	:param fromfiledate:
	:param tofiledate:
	:param n:
	:param lineterm:
	:param removed_colour: The :class:`~consolekit.terminal_colours.Colour` to use for lines that were removed.
	:param added_colour: The :class:`~consolekit.terminal_colours.Colour` to use for lines that were added.
	"""

	buf = StringList()
	diff = difflib.unified_diff(a, b, fromfile, tofile, fromfiledate, tofiledate, n, lineterm)

	for line in diff:
		if line.startswith('+'):
			buf.append(added_colour(line))
		elif line.startswith('-'):
			buf.append(removed_colour(line))
		else:
			buf.append(line)

	buf.blankline(ensure_single=True)

	return str(buf)


solidus_spinner = cycle("|/-\\")
"""
:func:`itertools.cycle` of characters to use as a loading spinner.

.. versionadded:: 0.7.0
"""

braille_spinner = cycle(['⢿', '⣻', '⣽', '⣾', '⣷', '⣯', '⣟', '⡿'])
"""
:func:`itertools.cycle` of braille characters to use as a loading spinner.

.. versionadded:: 0.7.0
"""

snake_spinner = cycle(['⠋', '⠙', '⠸', '⠴', '⠦', '⠇'])
"""
:func:`itertools.cycle` of braille characters to use as a loading spinner which looks like a snake.

.. versionadded:: 1.1.0
"""


def hide_cursor() -> None:
	"""
	Hide the cursor.

	To show it again use :func:`~.show_cursor`,
	or use the :func:`~.hidden_cursor` context manager.

	.. versionadded:: 0.7.0
	"""

	click.echo(
			terminal_colours.Cursor.HIDE(),
			nl=False,
			color=terminal_colours.resolve_color_default(),
			)


def show_cursor() -> None:
	"""
	Show the cursor.

	.. seealso:: The  :func:`~.hidden_cursor` context manager.

	.. versionadded:: 0.7.0
	"""

	click.echo(
			terminal_colours.Cursor.SHOW(),
			nl=False,
			color=terminal_colours.resolve_color_default(),
			)


@contextlib.contextmanager
def hidden_cursor() -> Iterator:
	"""
	Context manager to hide the cursor for the scope of the ``with`` block.

	.. versionadded:: 0.7.0

	.. versionchanged:: 0.9.0  Moved to :mod:`consolekit.utils`.
	"""

	try:
		hide_cursor()
		yield
	finally:
		show_cursor()


@lru_cache(1)
def _pycharm_terminal() -> bool:
	try:
		# 3rd party
		import psutil  # type: ignore[import]  # nodep

		parent_process = psutil.Process(os.getppid())
		grandparent_process = psutil.Process(parent_process.ppid())
		great_grandparent_process = psutil.Process(grandparent_process.ppid())
		great_grandparent_name = great_grandparent_process.name()

		#: TODO: pycharm on Windows and macOS
		return great_grandparent_name == "pycharm.sh"

	except Exception:  # pragma: no cover
		return False


class TerminalRenderer(BaseRenderer):
	"""
	Mistletoe markdown renderer for terminals.

	Tested in Gnome Terminal and Terminator (both libVTE-based), and PyCharm.
	libVTE has the best support.
	PyCharm's support for italics and strikethrough is poor.
	Support on Windows is, as expected, poor.

	Not tested on other terminals, but contributions are welcome to improve support.

	.. versionadded:: 0.8.0
	"""

	def render_strong(self, token: span_token.Strong) -> str:
		"""
		Render strong (``**strong**``).

		:param token: The token to render.
		"""

		return terminal_colours.Style.BRIGHT(self.render_inner(token))

	def render_emphasis(self, token: span_token.Emphasis) -> str:
		"""
		Render emphasis (``*emphasis*``).

		:param token: The token to render.
		"""

		if int(os.environ.get("PYCHARM_HOSTED", 0)):
			# Pycharm terminal doesn't support italic escape
			return SANS_SERIF_ITALIC_LETTERS(self.render_inner(token))

		return ''.join([
				terminal_colours.code_to_chars(3),
				self.render_inner(token),
				terminal_colours.code_to_chars(23),
				])

	def render_inline_code(self, token: span_token.InlineCode) -> str:
		r"""
		Render inline code (:inline-code:`code`).

		:param token: The token to render.
		"""

		# TODO: A better implementation
		return f"{self.render_inner(token)!r}"

	def render_strikethrough(self, token: span_token.Strikethrough) -> str:
		"""
		Render strikethrough (``~~strikethrough~~``).

		:param token: The token to render.
		"""

		if int(os.environ.get("PYCHARM_HOSTED", 0)) or _pycharm_terminal():
			# Pycharm terminal doesn't support strikethrough
			return self.render_inner(token)

		return ''.join([f'{char}\u0336' for char in self.render_inner(token)])

	def render_paragraph(self, token: block_token.Paragraph) -> str:
		"""
		Render a paragraph.

		:param token: The token to render.
		"""

		return f'\n{self.render_inner(token)}\n'

	_in_ordered_list: bool = False
	_ol_number: int = 0

	def render_list(self, token: block_token.List) -> str:
		"""
		Render a markdown list.

		:param token: The token to render.
		"""

		self._in_ordered_list = token.start is not None
		if self._in_ordered_list:
			self._ol_number = 0

		return self.render_inner(token) + '\n'

	def render_list_item(self, token: block_token.ListItem) -> str:
		"""
		Render a markdown list item.

		:param token:
		"""

		if self._in_ordered_list:
			self._ol_number += 1
			return f" {self._ol_number}. {self.render_inner(token).lstrip()}"

		else:
			return f" * {self.render_inner(token).lstrip()}"

	@staticmethod
	def render_line_break(token) -> str:  # noqa: MAN001
		"""
		Render a line break in a multiline paragraph.

		:param token:
		"""

		return ' '

	def render(self, token) -> str:  # noqa: MAN001
		"""
		Render the given token for display in a terminal.

		:param token:
		"""

		return super().render(token).replace("\n\n\n", "\n\n")


def long_echo(
		text: Union[str, StringList, Iterable[str]],
		use_pager: Optional[bool] = None,
		colour: ColourTrilean = None
		) -> None:
	"""
	Echo ``text`` to the terminal, optionally via a pager.

	.. versionadded:: 1.2.0

	:param text:
	:param use_pager: If :py:obj:`True`, forces the use of the pager. If :py:obj:`False` the pager is never used.
		If :py:obj:`None` the pager is used if `sys.stdout`` is a TTY and the number of lines is
		less than the terminal height.
	:param colour:  Whether to use coloured output. Default auto-detect.
	:no-default colour:

	.. tip:: Allow the user to control the value of ``use_pager`` with the :func:`no_pager_option` decorator.
	.. latex:clearpage::
	"""

	if isinstance(text, str):
		text = StringList(text.split('\n'))
	elif not isinstance(text, StringList):
		text = StringList(text)

	if use_pager is None:
		use_pager = True

		if shutil.get_terminal_size().lines >= len(text):
			# Don't use pager if fewer lines than terminal height
			use_pager = False

		if not sys.stdout.isatty():
			use_pager = False

	if use_pager:
		return click.echo_via_pager(str(text), color=colour)
	else:
		return click.echo(str(text), color=colour)



# stdlib
from collections import deque

# 3rd party
from coincidence.regressions import AdvancedFileRegressionFixture, check_file_regression

# this package
from consolekit import terminal_colours
from consolekit.terminal_colours import Colour, print_256_colour_testpattern, strip_ansi


def test_terminal_colours_constants():
	assert terminal_colours.CSI == "\u001b["
	assert terminal_colours.OSC == "\u001b]"
	assert terminal_colours.BEL == '\x07'


def test_terminal_colours_stacks():
	assert terminal_colours.fore_stack == deque([terminal_colours.Fore.RESET])
	assert terminal_colours.back_stack == deque([terminal_colours.Back.RESET])
	assert terminal_colours.style_stack == deque([terminal_colours.Style.NORMAL])


def test_terminal_colours_functions():
	assert terminal_colours.set_title("Foo") == "\u001b]2;Foo\u0007"

	# assert terminal_colours.clear_screen() == "\033[2J"
	# assert terminal_colours.clear_screen(1) == "\033[1J"

	assert terminal_colours.clear_line() == "\u001b[2K"
	assert terminal_colours.clear_line(1) == "\u001b[1K"


def test_ansi_cursor():
	assert terminal_colours.Cursor.UP() == "\u001b[1A"
	assert terminal_colours.Cursor.UP(1) == "\u001b[1A"
	assert terminal_colours.Cursor.UP(2) == "\u001b[2A"
	assert terminal_colours.Cursor.UP(3) == "\u001b[3A"

	assert terminal_colours.Cursor.DOWN() == "\u001b[1B"
	assert terminal_colours.Cursor.DOWN(1) == "\u001b[1B"
	assert terminal_colours.Cursor.DOWN(2) == "\u001b[2B"
	assert terminal_colours.Cursor.DOWN(3) == "\u001b[3B"

	assert terminal_colours.Cursor.FORWARD() == "\u001b[1C"
	assert terminal_colours.Cursor.FORWARD(1) == "\u001b[1C"
	assert terminal_colours.Cursor.FORWARD(2) == "\u001b[2C"
	assert terminal_colours.Cursor.FORWARD(3) == "\u001b[3C"

	assert terminal_colours.Cursor.BACK() == "\u001b[1D"
	assert terminal_colours.Cursor.BACK(1) == "\u001b[1D"
	assert terminal_colours.Cursor.BACK(2) == "\u001b[2D"
	assert terminal_colours.Cursor.BACK(3) == "\u001b[3D"

	assert terminal_colours.Cursor.POS() == "\u001b[1;1H"
	assert terminal_colours.Cursor.POS(1) == "\u001b[1;1H"
	assert terminal_colours.Cursor.POS(2) == "\u001b[1;2H"
	assert terminal_colours.Cursor.POS(3) == "\u001b[1;3H"
	assert terminal_colours.Cursor.POS(y=1) == "\u001b[1;1H"
	assert terminal_colours.Cursor.POS(y=2) == "\u001b[2;1H"
	assert terminal_colours.Cursor.POS(y=3) == "\u001b[3;1H"
	assert terminal_colours.Cursor.POS(x=1) == "\u001b[1;1H"
	assert terminal_colours.Cursor.POS(x=2) == "\u001b[1;2H"
	assert terminal_colours.Cursor.POS(x=3) == "\u001b[1;3H"
	assert terminal_colours.Cursor.POS(1, 1) == "\u001b[1;1H"
	assert terminal_colours.Cursor.POS(2, 2) == "\u001b[2;2H"
	assert terminal_colours.Cursor.POS(3, 3) == "\u001b[3;3H"
	assert terminal_colours.Cursor.POS(x=2, y=3) == "\u001b[3;2H"


class TestColour:

	def test_from_code(self):
		black = Colour.from_code(30)
		assert black == "\u001b[30m"
		assert black("Hello") == "\u001b[30mHello\u001b[39m"

		black = Colour.from_code(40, background=True)
		assert black == "\u001b[40m"
		assert black("Hello") == "\u001b[40mHello\u001b[49m"

	def test_from_rgb(self):
		yellow = Colour.from_rgb(252, 186, 3)
		assert yellow == "\u001b[38;2;252;186;3m"
		assert yellow("Hello") == "\u001b[38;2;252;186;3mHello\u001b[39m"

		yellow = Colour.from_rgb(252, 186, 3, background=True)
		assert yellow == "\u001b[48;2;252;186;3m"
		assert yellow("Hello") == "\u001b[48;2;252;186;3mHello\u001b[49m"

	def test_from_hex(self):
		yellow = Colour.from_hex("#fcba03")
		assert yellow == "\u001b[38;2;252;186;3m"
		assert yellow("Hello") == "\u001b[38;2;252;186;3mHello\u001b[39m"

		yellow = Colour.from_hex("#fcba03", background=True)
		assert yellow == "\u001b[48;2;252;186;3m"
		assert yellow("Hello") == "\u001b[48;2;252;186;3mHello\u001b[49m"


def test_print_256_colour_testpattern(
		monkeypatch,
		advanced_file_regression: AdvancedFileRegressionFixture,
		capsys,
		):
	monkeypatch.setattr(terminal_colours, "resolve_color_default", lambda *args: False)

	# Checks only the structure, not the colours
	print_256_colour_testpattern()

	check_file_regression(strip_ansi(capsys.readouterr().out.strip()), advanced_file_regression)



#!/usr/bin/env python3
#
#  __init__.py
"""
Additional utilities for `click <https://click.palletsprojects.com/en/7.x/>`_.

.. attention::

	``consolekit`` disables Python's readline history to prevent unrelated histories appearing for prompts.
	If the original behaviour is desired run:

	.. code-block:: python

		import readline
		readline.set_history_length(-1)
		readline.set_auto_history(True)

"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# pylint: disable=redefined-builtin

# stdlib
from typing import Any, Callable, Dict, Optional, Type, TypeVar, cast

# 3rd party
import click

# this package
from consolekit import _readline, commands, input, terminal_colours, tracebacks, utils  # noqa: F401
from consolekit.commands import SuggestionGroup
from consolekit.options import _Option

# pylint: enable=redefined-builtin

__author__: str = "Dominic Davis-Foster"
__copyright__: str = "2020 Dominic Davis-Foster"
__license__: str = "MIT License"
__version__: str = "1.7.2"
__email__: str = "dominic@davis-foster.co.uk"

__all__ = (
		"CONTEXT_SETTINGS",
		"click_command",
		"click_group",
		"option",
		"SuggestionGroup",
		)

_C = TypeVar("_C", bound=click.Command)
_G = TypeVar("_G", bound=click.Group)

CONTEXT_SETTINGS: Dict[str, Any] = dict(help_option_names=["-h", "--help"], max_content_width=120)


def click_command(
		name: Optional[str] = None,
		cls: Optional[Type[_C]] = None,
		**attrs: Any,
		) -> Callable[[Callable], _C]:
	r"""
	Shortcut to :func:`click.command`, with the ``-h``/``--help`` option enabled and a max width of ``120``.

	:param name:
	:param cls:
	:type cls: :class:`~typing.Type`\[:class:`~click.Command`\]
	:default type: :class:`click.Command`
	:param \*\*attrs: Additional keyword arguments passed to the :class:`~click.Command`.

	:rtype: :class:`~typing.Callable`\[[:class:`~typing.Callable`\], :class:`~click.Command`\]
	"""

	attrs.setdefault("context_settings", CONTEXT_SETTINGS)
	return cast(Callable[[Callable], _C], click.command(name, cls=cls, **attrs))


def click_group(
		name: Optional[str] = None,
		cls: Optional[Type[_G]] = None,
		**attrs: Any,
		) -> Callable[[Callable], _G]:
	r"""
	Shortcut to :func:`click.group`, with the ``-h``/``--help`` option enabled and a max width of ``120``.

	:param name:
	:param cls:
	:type cls: :class:`~typing.Type`\[:class:`~click.Group`\]
	:default type: :class:`click.Group`
	:param \*\*attrs: Additional keyword arguments passed to the :class:`~click.Group`.

	:rtype:

	.. latex:clearpage::
	"""

	if cls is None:
		cls = SuggestionGroup

	attrs.setdefault("context_settings", CONTEXT_SETTINGS)
	return click_command(name, cls=cls, **attrs)


def option(
		*param_decls: str,
		**attrs: Any,
		) -> Callable[[_C], _C]:
	r"""
	Shortcut to :func:`click.option`, but using :func:`consolekit.input.confirm` when prompting for a boolean flag.

	:param \*param_decls:
	:param \*\*attrs: Additional keyword arguments passed to :func:`click.command`.
	"""

	attrs.setdefault("cls", _Option)
	return cast(Callable[[_C], _C], click.option(*param_decls, **attrs))


# Fixes intersphinx links
click.Command.__module__ = "click"
click.Argument.__module__ = "click"
click.Abort.__module__ = "click"
click.Option.__module__ = "click"
click.ParamType.__module__ = "click"
click.Parameter.__module__ = "click"
click.Context.__module__ = "click"
click.HelpFormatter.__module__ = "click"
click.Group.__module__ = "click"
click.OptionParser.__module__ = "click"



# stdlib
import re
from typing import Callable, ContextManager, Type

# 3rd party
import click
import pytest
from coincidence.regressions import AdvancedFileRegressionFixture
from pytest_regressions.file_regression import FileRegressionFixture

# this package
from consolekit import click_command
from consolekit.testing import CliRunner, Result, _click_major
from consolekit.tracebacks import TracebackHandler, handle_tracebacks, traceback_handler, traceback_option

exceptions = pytest.mark.parametrize(
		"exception",
		[
				pytest.param(FileNotFoundError("foo.txt"), id="FileNotFoundError"),
				pytest.param(FileExistsError("foo.txt"), id="FileExistsError"),
				pytest.param(Exception("Something's awry!"), id="Exception"),
				pytest.param(ValueError("'age' must be >= 0"), id="ValueError"),
				pytest.param(TypeError("Expected type int, got type str"), id="TypeError"),
				pytest.param(NameError("name 'hello' is not defined"), id="NameError"),
				pytest.param(SyntaxError("invalid syntax"), id="SyntaxError"),
				]
		)
contextmanagers = pytest.mark.parametrize(
		"contextmanager, exit_code",
		[
				pytest.param(handle_tracebacks, 1, id="handle_tracebacks"),
				pytest.param(traceback_handler, 1, id="traceback_handler"),
				pytest.param(TracebackHandler(), 1, id="TracebackHandler"),
				pytest.param(TracebackHandler(SystemExit(1)), 1, id="TracebackHandler_SystemExit"),
				pytest.param(TracebackHandler(SystemExit(2)), 2, id="TracebackHandler_SystemExit_2"),
				]
		)


@exceptions
@contextmanagers
def test_handle_tracebacks(
		exception: BaseException,
		contextmanager: Callable[..., ContextManager],
		exit_code: int,
		advanced_file_regression: AdvancedFileRegressionFixture,
		cli_runner: CliRunner,
		):

	@click.command()
	def demo() -> None:

		with contextmanager():
			raise exception

	result: Result = cli_runner.invoke(demo, catch_exceptions=False)
	result.check_stdout(advanced_file_regression)
	assert result.exit_code == exit_code


@exceptions
def test_handle_tracebacks_show_traceback(exception: Exception, cli_runner: CliRunner):

	@click.command()
	def demo() -> None:

		with handle_tracebacks(show_traceback=True):
			raise exception

	with pytest.raises(type(exception), match=re.escape(str(exception))):
		cli_runner.invoke(demo, catch_exceptions=False)


@pytest.mark.parametrize("exception", [EOFError(), KeyboardInterrupt(), click.Abort()])
@contextmanagers
def test_handle_tracebacks_ignored_exceptions_click(
		exception: Exception,
		contextmanager: Callable[..., ContextManager],
		cli_runner: CliRunner,
		exit_code: int,
		):

	@click.command()
	def demo() -> None:

		with contextmanager():
			raise exception

	result: Result = cli_runner.invoke(demo, catch_exceptions=False)

	assert result.stdout.strip() == "Aborted!"
	assert result.exit_code == 1


@pytest.mark.parametrize("exception", [EOFError, KeyboardInterrupt, click.Abort, SystemExit])
@contextmanagers
def test_handle_tracebacks_ignored_exceptions(
		exception: Type[Exception],
		contextmanager: Callable[..., ContextManager],
		exit_code: int,
		):

	with pytest.raises(exception):  # noqa: PT012  # skipcq
		with contextmanager():
			raise exception


@pytest.mark.parametrize(
		"exception, code",
		[
				pytest.param(click.UsageError("Message"), 2, id="click.UsageError"),
				pytest.param(click.BadParameter("Message"), 2, id="click.BadParameter"),
				pytest.param(
						click.FileError("Message"),
						1,
						id="click.FileError",
						marks=pytest.mark.skipif(_click_major == 8, reason="Output differs on Click 8")
						),
				pytest.param(
						click.FileError("Message"),
						1,
						id="click.FileError_8",
						marks=pytest.mark.skipif(_click_major != 8, reason="Output differs on Click 8")
						),
				pytest.param(click.ClickException("Message"), 1, id="click.ClickException"),
				]
		)
@contextmanagers
@pytest.mark.parametrize(
		"click_version",
		[
				pytest.param(
						'7',
						marks=pytest.mark.skipif(_click_major == 8, reason="Output differs on click 8"),
						),
				pytest.param(
						'8',
						marks=pytest.mark.skipif(_click_major != 8, reason="Output differs on click 8"),
						),
				]
		)
def test_handle_tracebacks_ignored_click(
		exception: Exception,
		contextmanager: Callable[..., ContextManager],
		advanced_file_regression: AdvancedFileRegressionFixture,
		code: int,
		cli_runner: CliRunner,
		click_version: str,
		exit_code: int,
		):

	@click.command()
	def demo() -> None:

		with contextmanager():
			raise exception

	result = cli_runner.invoke(demo, catch_exceptions=False)
	result.check_stdout(advanced_file_regression)

	# if code == 1 and exit_code == 2:
	# 	code = 2

	assert result.exit_code == code


def test_traceback_option(file_regression: FileRegressionFixture, cli_runner: CliRunner):

	@traceback_option()
	@click_command()
	def main(show_traceback: bool) -> None:
		print(show_traceback)

	result = cli_runner.invoke(main, catch_exceptions=False)
	assert result.stdout.rstrip() == "False"
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args="--traceback")
	assert result.stdout.rstrip() == "True"
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args="-T")
	assert result.stdout.rstrip() == "True"
	assert result.exit_code == 0

	result = cli_runner.invoke(main, args="-h")
	result.check_stdout(file_regression)
	assert result.exit_code == 0


def test_traceback_handler_abort(capsys):

	TH = TracebackHandler(ValueError("foo"))

	with pytest.raises(ValueError, match="foo$"):
		TH.abort("Hello World")

	assert capsys.readouterr().err == "Hello World\n"

	with pytest.raises(ValueError, match="foo$"):
		TH.abort(["Hello", "Everybody"])

	assert capsys.readouterr().err == "HelloEverybody\n"



#!/usr/bin/env python3
#
#  commands.py
"""
Customised click commands and command groups.

.. versionadded:: 0.8.0
"""
#
#  Copyright © 2020-2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  MarkdownHelpCommand.parse_args and the docstrings of ContextInheritingGroup
#  based on https://github.com/pallets/click
#  Copyright 2014 Pallets
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |      * Redistributions of source code must retain the above copyright notice,
#  |        this list of conditions and the following disclaimer.
#  |      * Redistributions in binary form must reproduce the above copyright notice,
#  |        this list of conditions and the following disclaimer in the documentation
#  |        and/or other materials provided with the distribution.
#  |      * Neither the name of the copyright holder nor the names of its contributors
#  |        may be used to endorse or promote products derived from this software without
#  |        specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  |  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  |  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  |  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
#  |  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
#  |  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#  |  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
#  |  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
#  |  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
#  |  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# stdlib
import difflib
import inspect
from textwrap import indent
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Tuple

# 3rd party
import click
from click.core import iter_params_for_processing
from click.parser import split_opt
from click.utils import make_str
from domdf_python_tools.stringlist import DelimitedList
from domdf_python_tools.words import Plural
from mistletoe import block_token  # type: ignore[import]

# this package
from consolekit.terminal_colours import ColourTrilean, resolve_color_default, strip_ansi
from consolekit.utils import TerminalRenderer

__all__ = (
		"MarkdownHelpCommand",
		"MarkdownHelpGroup",
		"MarkdownHelpMixin",
		"RawHelpCommand",
		"RawHelpGroup",
		"RawHelpMixin",
		"SuggestionGroup",
		"ContextInheritingGroup",
		)

_argument = Plural("argument", "arguments")


class RawHelpMixin:
	"""
	Mixin class for :class:`click.Command` and :class:`click.Group` which leaves the help text unformatted.

	.. seealso:: :class:`~.RawHelpCommand` and :class:`~.RawHelpGroup`

	.. tip:: This can be combined with groups such as :class:`~.SuggestionGroup`.

	.. versionadded:: 0.8.0
	"""

	help: Optional[str]  # noqa: A003  # pylint: disable=redefined-builtin

	def format_help_text(self, ctx: click.Context, formatter: click.formatting.HelpFormatter) -> None:
		"""
		Writes the help text to the formatter if it exists.

		:param ctx:
		:param formatter:
		"""

		if not self.help:
			return

		text = inspect.cleandoc(self.help or '').partition('\x0c')[0]

		if getattr(self, "deprecated", False):
			text = f"(Deprecated) {text}"

		formatter.write('\n')
		formatter.write(indent(text, "  "))
		formatter.write('\n')


class RawHelpCommand(RawHelpMixin, click.Command):  # lgtm [py/conflicting-attributes]
	"""
	Subclass of :class:`click.Command` which leaves the help text unformatted.

	.. versionadded:: 0.8.0
	"""


class RawHelpGroup(RawHelpMixin, click.Group):  # lgtm [py/conflicting-attributes]
	"""
	Subclass of :class:`click.Group` which leaves the help text unformatted.

	.. versionadded:: 0.8.0
	"""


class MarkdownHelpMixin:
	"""
	Mixin class for :class:`click.Command` and :class:`click.Group` which treats the help text as markdown
	and prints a rendered representation.

	.. seealso:: :class:`~.MarkdownHelpCommand` and :class:`~.MarkdownHelpGroup`

	.. tip:: This can be combined with groups such as :class:`~.SuggestionGroup`.

	Tested in Gnome Terminal and Terminator (both libVTE-based), and PyCharm.
	libVTE has the best support.
	PyCharm's support for italics and strikethrough is poor.
	Support on Windows is, as expected, poor.

	Not tested on other terminals, but contributions are welcome to improve support.

	.. versionadded:: 0.8.0
	"""  # noqa: D400

	help: Optional[str]  # noqa: A003  # pylint: disable=redefined-builtin
	no_args_is_help: bool
	_colour: ColourTrilean = None

	if TYPE_CHECKING:

		def get_params(self, ctx: click.Context) -> List[click.Parameter]:
			"""
			Returns the list of parameters for this command in the order they should appear in the help page.

			:param ctx:
			"""

			raise NotImplementedError

		def make_parser(self, ctx: click.Context) -> click.OptionParser:
			"""
			Creates the underlying option parser for this command.

			:param ctx:
			"""

			raise NotImplementedError

	def format_help_text(self, ctx: click.Context, formatter: click.formatting.HelpFormatter) -> None:
		"""
		Writes the help text to the formatter if it exists.

		:param ctx:
		:param formatter:
		"""

		if not self.help:
			return

		text = inspect.cleandoc(self.help or '').partition('\x0c')[0]

		if getattr(self, "deprecated", False):
			text = f"(Deprecated) {text}"

		doc = block_token.Document(text)

		with TerminalRenderer() as renderer:
			rendered_doc = indent(renderer.render(doc).strip(), "  ")

		if resolve_color_default(self._colour) is False:
			# Also remove 'COMBINING LONG STROKE OVERLAY', used for strikethrough.
			rendered_doc = strip_ansi(rendered_doc).replace('̶', '')

		formatter.write('\n')
		formatter.write(rendered_doc)
		formatter.write('\n')


def parse_help_args(obj: MarkdownHelpMixin, ctx: click.Context, args: List[str]) -> List[str]:
	"""
	Helper function for markdown help classes to parse the given arguments
	and modify the context as necessary.

	.. versionadded:: 1.1.0

	:param obj:
	:param ctx:
	:param args:
	"""  # noqa: D400

	# This is necessary to parse any --colour/--no-colour commands before generating the help,
	# to ensure the option is honoured.

	if not args and obj.no_args_is_help and not ctx.resilient_parsing:
		click.echo(ctx.get_help(), color=ctx.color)
		ctx.exit()

	parser = obj.make_parser(ctx)
	opts, args, param_order = parser.parse_args(args=args)

	obj._colour = opts.get("colour", ctx.color)

	for param in iter_params_for_processing(param_order, obj.get_params(ctx)):
		value, args = param.handle_parse_result(ctx, opts, args)

	if args and not ctx.allow_extra_args and not ctx.resilient_parsing:
		args_string = DelimitedList(map(make_str, args))
		ctx.fail(f"Got unexpected extra {_argument(len(args))} ({args_string: })")

	ctx.args = args
	return args


class MarkdownHelpCommand(MarkdownHelpMixin, click.Command):  # lgtm [py/conflicting-attributes]
	"""
	Subclass of :class:`click.Command` which treats the help text as markdown
	and prints a rendered representation.

	Tested in Gnome Terminal and Terminator (both libVTE-based), and PyCharm.
	libVTE has the best support.
	PyCharm's support for italics and strikethrough is poor.
	Support on Windows is, as expected, poor.

	Not tested on other terminals, but contributions are welcome to improve support.

	.. versionadded:: 0.8.0
	"""  # noqa: D400

	def parse_args(self, ctx: click.Context, args: List[str]) -> List[str]:
		"""
		Parse the given arguments and modify the context as necessary.

		:param ctx:
		:param args:
		"""

		return parse_help_args(self, ctx, args)


class MarkdownHelpGroup(MarkdownHelpMixin, click.Group):  # lgtm [py/conflicting-attributes]
	"""
	Subclass of :class:`click.Group` which treats the help text as markdown
	and prints a rendered representation.

	Tested in Gnome Terminal and Terminator (both libVTE-based), and PyCharm.
	libVTE has the best support.
	PyCharm's support for italics and strikethrough is poor.
	Support on Windows is, as expected, poor.

	Not tested on other terminals, but contributions are welcome to improve support.

	.. versionadded:: 0.8.0
	"""  # noqa: D400

	def parse_args(self, ctx: click.Context, args: List[str]) -> List[str]:
		"""
		Parse the given arguments and modify the context as necessary.

		:param ctx:
		:param args:
		"""

		# This is necessary to parse any --colour/--no-colour commands before generating the help,
		# to ensure the option is honoured.

		if not args and self.no_args_is_help and not ctx.resilient_parsing:
			click.echo(ctx.get_help(), color=ctx.color)
			ctx.exit()

		rest = parse_help_args(self, ctx, args)

		if self.chain:
			ctx.protected_args = rest
			ctx.args = []
		elif rest:
			ctx.protected_args, ctx.args = rest[:1], rest[1:]

		return ctx.args


class ContextInheritingGroup(click.Group):
	"""
	Subclass of :class:`click.Group` whose children inherit its :attr:`~click.BaseCommand.context_settings`.

	The group's commands can be given different context settings by passing the
	``context_settings`` keyword argument to :meth:`~.command` and :meth:`~.group` as normal.

	.. versionadded:: 1.1.0

	.. autosummary-widths:: 5/16
	"""

	def command(self, *args, **kwargs) -> Callable[[Callable[..., Any]], click.Command]:  # type: ignore[override]
		"""
		A shortcut decorator for declaring and attaching a command to the group.

		This takes the same arguments as :func:`click.command` but
		immediately registers the created command with this instance
		by calling into :meth:`click.Group.add_command`.
		"""

		kwargs.setdefault("context_settings", self.context_settings)
		return super().command(*args, **kwargs)

	def group(self, *args, **kwargs) -> Callable[[Callable[..., Any]], click.Group]:  # type: ignore[override]
		"""
		A shortcut decorator for declaring and attaching a group to the group.

		This takes the same arguments as :func:`click.group` but
		immediately registers the created group with this instance
		by calling into :meth:`click.Group.add_command`.
		"""

		kwargs.setdefault("context_settings", self.context_settings)
		return super().group(*args, **kwargs)


class SuggestionGroup(ContextInheritingGroup):
	"""
	Subclass of :class:`click.Group` which suggests the most similar command if the command is not found.

	.. versionadded 0.2.0

	.. versionchanged:: 0.8.0

		Moved to :mod:`consolekit.commands`.

	.. versionchanged:: 1.1.0

		Now inherits from :class:`~.ContextInheritingGroup`
	"""

	def resolve_command(
			self,
			ctx: click.Context,
			args: List[str],
			) -> Tuple[str, click.Command, List[str]]:
		"""
		Resolve the requested command belonging to this group, and print a suggestion if it can't be found.

		:param ctx:
		:param args:

		:return: The name of the matching command,
			the :class:`click.Command` object itself,
			and any remaining arguments.
		"""

		cmd_name = make_str(args[0])
		original_cmd_name = cmd_name

		# Get the command
		cmd = self.get_command(ctx, cmd_name)

		# If we can't find the command but there is a normalization
		# function available, we try with that one.
		if cmd is None and ctx.token_normalize_func is not None:
			cmd_name = ctx.token_normalize_func(cmd_name)
			cmd = self.get_command(ctx, cmd_name)

		# If we don't find the command we want to show an error message
		# to the user that it was not provided.
		# However, there is something else we should do:
		# if the first argument looks like an option we want to kick off parsing again
		# for arguments to resolve things like --help which now should go to the main place.
		if cmd is None and not ctx.resilient_parsing:
			if split_opt(cmd_name)[0]:
				self.parse_args(ctx, ctx.args)

			closest = difflib.get_close_matches(original_cmd_name, self.commands, n=1)
			message = [f"No such command '{original_cmd_name}'."]
			if closest:
				message.append(f"The most similar command is {closest[0]!r}.")
			ctx.fail('\n'.join(message))

		# TODO: cmd here is Optional[click.Command], typeshed says it should be just click.Command
		#  I think typeshed is wrong.
		#  https://github.com/python/typeshed/blob/484c014665cdf071b292dd9630f207c03e111895/third_party/2and3/click/core.pyi#L171
		return cmd_name, cmd, args[1:]  # type: ignore[return-value]



# stdlib
from typing import Type

# 3rd party
import click
import pytest
from click import echo
from coincidence.regressions import AdvancedDataRegressionFixture
from domdf_python_tools.stringlist import StringList

# this package
from consolekit.input import choice, confirm, prompt
from consolekit.testing import _click_major


def test_choice_letters(
		capsys,
		monkeypatch,
		advanced_data_regression: AdvancedDataRegressionFixture,
		):

	inputs = iter(['F', 'G', 'D'])

	def fake_input(prompt: str) -> str:
		value = next(inputs)
		print(f"{prompt}{value}".rstrip())
		return value

	monkeypatch.setattr(click.termui, "visible_prompt_func", fake_input)

	echo("Configuration file '/etc/sudoers'")
	echo("==> Modified (by you or by a script) since installation.")
	echo("==> Package distributor has shipped an updated version.")
	echo("What would you like to do about it ?  Your options are:")
	options = {
			'Y': "install the package maintainer's version",
			'N': "keep your currently-installed version",
			'D': "show the differences between the versions",
			'Z': "start a shell to examine the situation"
			}
	assert choice(text="*** sudoers", options=options, default='N') == 'D'

	advanced_data_regression.check(list(StringList(capsys.readouterr().out.splitlines())))


@pytest.mark.parametrize(
		"click_version",
		[
				pytest.param(
						'7',
						marks=pytest.mark.skipif(_click_major == 8, reason="Output differs on click 8"),
						),
				pytest.param(
						'8',
						marks=pytest.mark.skipif(_click_major != 8, reason="Output differs on click 8"),
						),
				]
		)
def test_choice_numbers(
		capsys,
		monkeypatch,
		advanced_data_regression: AdvancedDataRegressionFixture,
		click_version: str,
		):

	inputs = iter(["20", '0', '5'])

	def fake_input(prompt: str) -> str:
		value = next(inputs)
		print(f"{prompt}{value}".rstrip())
		return value

	monkeypatch.setattr(click.termui, "visible_prompt_func", fake_input)

	echo("What is the Development Status of this project?")
	options = [
			"Planning",
			"Pre-Alpha",
			"Alpha",
			"Beta",
			"Production/Stable",
			"Mature",
			"Inactive",
			]
	assert choice(text='', options=options, start_index=1) == 4

	advanced_data_regression.check(list(StringList(capsys.readouterr().out.splitlines())))


def test_confirm(
		capsys,
		monkeypatch,
		advanced_data_regression: AdvancedDataRegressionFixture,
		):

	inputs = iter(['Y', 'N', '', '', "yEs", "No", "gkjhkhjv", ''])

	def fake_input(prompt: str) -> str:
		value = next(inputs)
		print(f"{prompt}{value}".rstrip())
		return value

	monkeypatch.setattr(click.termui, "visible_prompt_func", fake_input)

	assert confirm(text="Do you wish to delete all files in '/' ?", default=False) is True
	assert confirm(text="Do you wish to delete all files in '/' ?", default=False) is False
	assert confirm(text="Do you wish to delete all files in '/' ?", default=False) is False
	assert confirm(text="Do you wish to delete all files in '/' ?", default=True) is True
	assert confirm(text="Do you wish to delete all files in '/' ?", default=False) is True
	assert confirm(text="Do you wish to delete all files in '/' ?", default=True) is False
	assert confirm(text="Do you wish to delete all files in '/' ?", default=True) is True

	advanced_data_regression.check(list(StringList(capsys.readouterr().out.splitlines())))


def test_prompt(
		capsys,
		monkeypatch,
		advanced_data_regression: AdvancedDataRegressionFixture,
		):

	inputs = iter([
			'',
			'',
			'',
			'',
			"24",
			"Bond007",
			"badpassword",
			"baspassword",
			"badpassword",
			"badpassword",
			"badpassword",
			"badpassword",
			])

	def fake_input(prompt: str) -> str:
		value = next(inputs)
		print(f"{prompt}{value}".rstrip())
		return value

	monkeypatch.setattr(click.termui, "visible_prompt_func", fake_input)

	assert prompt(text="What is your age", prompt_suffix="? ", type=click.INT) == 24

	assert prompt(text="Username", type=click.STRING) == "Bond007"
	assert prompt(text="Password", type=click.STRING, confirmation_prompt=True) == "badpassword"
	assert prompt(
			text="Password",
			type=click.STRING,
			confirmation_prompt="Are you sure about that? ",
			) == "badpassword"

	advanced_data_regression.check(list(StringList(capsys.readouterr().out.splitlines())))


@pytest.mark.parametrize("exception", [KeyboardInterrupt, EOFError])
def test_prompt_abort(
		capsys,
		monkeypatch,
		advanced_data_regression: AdvancedDataRegressionFixture,
		exception: Type[Exception],
		):

	def fake_input(prompt: str) -> str:
		print(f"{prompt}", end='')
		raise exception

	monkeypatch.setattr(click.termui, "visible_prompt_func", fake_input)

	with pytest.raises(click.Abort, match="^$"):
		prompt(text="Password", type=click.STRING, confirmation_prompt=True)

	assert list(StringList(capsys.readouterr().out.splitlines())) == ["Password:"]


@pytest.mark.parametrize("exception", [KeyboardInterrupt, EOFError])
def test_confirm_abort(
		capsys,
		monkeypatch,
		advanced_data_regression: AdvancedDataRegressionFixture,
		exception: Type[Exception],
		):

	def fake_input(prompt: str) -> str:
		print(f"{prompt}", end='')
		raise exception

	monkeypatch.setattr(click.termui, "visible_prompt_func", fake_input)

	with pytest.raises(click.Abort, match="^$"):
		confirm(text="Do you wish to delete all files in '/' ?", default=False)

	expected = ["Do you wish to delete all files in '/' ? [y/N]:"]
	assert list(StringList(capsys.readouterr().out.splitlines())) == expected



# 3rd party
import click

# this package
from consolekit import click_command
from consolekit.options import DescribedArgument
from consolekit.testing import CliRunner


def test_result(cli_runner: CliRunner):

	@click.argument(
			"dest",
			cls=DescribedArgument,
			type=click.STRING,
			description="The destination directory.",
			)
	@click_command()
	def main(dest: str) -> None:
		print(dest)

	result = cli_runner.invoke(main, catch_exceptions=False, args="./staging")
	assert result.stdout.rstrip() == "./staging"
	assert result.output.rstrip() == "./staging"
	assert result.exit_code == 0


def test_result_no_mix_stderr():

	cli_runner = CliRunner(mix_stderr=False)

	@click.argument(
			"dest",
			cls=DescribedArgument,
			type=click.STRING,
			description="The destination directory.",
			)
	@click_command()
	def main(dest: str) -> None:
		print(dest)

	result = cli_runner.invoke(main, catch_exceptions=False, args="./staging")
	assert result.stdout.rstrip() == "./staging"
	assert result.stderr == ''
	assert result.output.rstrip() == "./staging"
	assert result.exit_code == 0



# 3rd party
from docutils import nodes
from docutils.transforms.misc import Transitions
from sphinx import addnodes
from sphinx.application import Sphinx
from sphinx.config import Config
from sphinx.util.docutils import SphinxDirective, SphinxRole
from sphinx.util.nodes import process_only_nodes
from sphinx.writers.latex import LaTeXTranslator


class TransitionTransform(Transitions):
	default_priority = Transitions.default_priority - 10

	def visit_transition(self, node):
		index = node.parent.index(node)
		error = None
		if (
				index == 0 or isinstance(node.parent[0], nodes.title) and
				(index == 1 or isinstance(node.parent[1], nodes.subtitle) and index == 2)
				):
			if isinstance(node.parent, (nodes.document, nodes.section)):
				error = self.document.reporter.error(
						"Document or section may not begin with a transition.",
						source=node.source,
						line=node.line,
						)
		elif isinstance(node.parent[index - 1], nodes.transition):
			error = self.document.reporter.error(
					'At least one body element must separate transitions; '
					'adjacent transitions are not allowed.',
					source=node.source,
					line=node.line
					)

		if error:
			# Insert before node and update index.
			node.parent.insert(index, error)
			index += 1

		assert index < len(node.parent)  # skipcq

		if index != len(node.parent) - 1:
			# No need to move the node.
			return

		# Node behind which the transition is to be moved.
		sibling = node

		# While sibling is the last node of its parent.
		while index == len(sibling.parent) - 1:
			sibling = sibling.parent
			# If sibling is the whole document (i.e. it has no parent).
			if sibling.parent is None:
				# Transition at the end of document.  Do not move the
				# transition up, and place an error behind.
				error = self.document.reporter.error("Document may not end with a transition.", line=node.line)
				node.parent.insert(node.parent.index(node) + 1, error)
				return

			index = sibling.parent.index(sibling)

		# Remove the original transition node.
		node.parent.remove(node)
		# Insert the transition after the sibling.
		sibling.parent.insert(index + 1, node)


class TransitionTransform(Transitions):
	default_priority = Transitions.default_priority - 10

	def visit_transition(self, node):

		if isinstance(node.parent, addnodes.only):
			process_only_nodes(
					node.parent.parent,
					tags=self.document.settings.env.app.builder.tags,
					)


class TransitionDirective(SphinxDirective):

	def run(self):
		return [nodes.transition()]


def depart_desc_annotation(translator: LaTeXTranslator, node: addnodes.desc_annotation) -> None:
	translator.body.append("}}")
	parent = node.parent
	grandparent = parent.parent
	if len(grandparent) > 1:
		aunt = grandparent[grandparent.index(parent) + 1]

		if isinstance(parent, addnodes.desc_signature) and isinstance(aunt, addnodes.desc_signature):
			translator.body.append(r"\vspace{5px}")


class InlineRole(SphinxRole):

	def run(self):
		return [nodes.literal('', f"`{self.text}`")], []


def setup(app: Sphinx):
	app.add_directive("transition", TransitionDirective)
	app.add_transform(TransitionTransform)
	app.add_node(
			addnodes.desc_annotation,
			latex=(LaTeXTranslator.visit_desc_annotation, depart_desc_annotation),
			override=True
			)
	app.add_role("inline-code", InlineRole())
